# coding: utf-8

"""本模块为gfworkflow中各个组件的测试用例
"""

from girlfriend.exception import (
    InvalidArgumentException,
    UnknownWitchToExecuteException
)
from girlfriend.plugin import Plugin, PluginManager
from girlfriend.testing.plugin import PluginMgrFixture
from girlfriend.testing import GirlFriendTestCase
from girlfriend.workflow.protocol import (
    End,
    OkEnd,
    AbstractListener
)
from girlfriend.workflow.gfworkflow import (
    Context,
    Job,
    Decision,
    Workflow
)


class ContextTestCase(GirlFriendTestCase):

    def test_get_and_set(self):
        ctx = Context({}, {}, None, None)
        ctx["a"] = 1
        self.assertEquals(ctx["a"], 1)
        self.assertEquals(ctx.get("a"), 1)
        self.assertIs(ctx["b"], None)


class JobTestCase(GirlFriendTestCase):

    def test_execute(self):
        """测试Job运行
        """
        plugin_mgr_fixture = PluginMgrFixture(None)
        self.useFixture(plugin_mgr_fixture)
        plugin_mgr = plugin_mgr_fixture.plugin_manager

        plugin = Plugin.wrap_function(
            "test_job", "test job", lambda ctx, a, b, c: a + b + c)
        plugin_mgr.register(plugin)
        plugin_mgr.sys_prepare(None, "test_job")

        def add_one(ctx, x):
            ctx["a"] += x

        plugin = Plugin.wrap_function("add_one", "add one", add_one)
        plugin_mgr.register(plugin)
        plugin_mgr.sys_prepare(None, "add_one")

        ctx = Context({}, {}, plugin_mgr, None)

        job = Job(
            "test_job",
            args=(1, 2, 3)
        )

        self.assertEquals(job.execute(ctx), 6)

        # 当上下文中的参数是列表时，上下文中的参数取代列表
        ctx = Context({}, {"test_job": (-1, -2, -3)}, plugin_mgr, None)
        self.assertEquals(job.execute(ctx), -6)

        # 测试使用上下文中的变量

        ctx = Context({}, {
            "test_job": ("$a", "$b", "$c")
        }, plugin_mgr, None)

        ctx["a"], ctx["b"], ctx["c"] = 5, 6, 7
        self.assertEquals(job.execute(ctx), 5 + 6 + 7)

        # 字符串
        ctx = Context({}, {
            "test_job": ("a", "b", "cd")
        }, plugin_mgr, None)
        self.assertEquals(job.execute(ctx), "abcd")

        # 字典参数
        job = Job(
            "test_job",
            args={
                "a": 1,
                "b": 2,
                "c": 3
            }
        )

        ctx = Context({}, {}, plugin_mgr, None)
        self.assertEquals(job.execute(ctx), 6)

        # 如果context中同样包含参数，那么基于原先参数执行update操作
        ctx = Context({}, {"test_job": {"a": -1}}, plugin_mgr, None)
        self.assertEquals(job.execute(ctx), 4)

        # 替换Context中的参数
        ctx = Context({}, {"test_job": {
            "a": "$a",
            "b": "$b",
            "c": "$c"
        }}, plugin_mgr, None)

        ctx["a"], ctx["b"], ctx["c"] = -1, -2, -3
        self.assertEquals(job.execute(ctx), -6)

        # 初始化参数类型与上下文参数类型不一致时，抛出InvalidArgumentException
        ctx = Context({}, {"test_job": ("a", "b", "c")}, plugin_mgr, None)
        self.failUnlessException(InvalidArgumentException, Job.execute,
                                 job, ctx)

        # 同时指定了插件和可执行逻辑时抛出异常
        self.failUnlessException(
            UnknownWitchToExecuteException,
            Job,
            "test_plugin",
            plugin="test_plugin",
            caller=lambda ctx: "hehe"
        )

        # 仅指定可执行逻辑时的情况
        job = Job("hehe", caller=lambda ctx, a, b, c: a * b * c)
        ctx = Context({}, {"hehe": (5, 2, 3)}, plugin_mgr, None)
        self.assertEquals(job.execute(ctx), 30)

        # generator args

        def gen_args(context):
            for i in xrange(10):
                yield i,
        ctx = Context({}, {}, plugin_mgr, None)
        ctx["a"] = 0

        job = Job("hehe", plugin="add_one", args=gen_args)
        job.execute(ctx)
        self.assertEquals(ctx["a"], sum(xrange(0, 10)))


class WorkflowTestCase(GirlFriendTestCase):

    def setUp(self):
        # 构建测试所需要的插件
        plugin_mgr = PluginManager()

        def add_one(ctx, num):
            return num + 1
        add_one_plugin = Plugin.wrap_function("add_one", "add one", add_one)
        plugin_mgr.register(add_one_plugin)

        def add_two(ctx, num):
            return num + 2
        add_two_plugin = Plugin.wrap_function("add_two", "add two", add_two)
        plugin_mgr.register(add_two_plugin)

        def add_three(ctx, num):
            value = num + 3
            ctx["add_three"] = value
        add_three_plugin = Plugin.wrap_function(
            "add_three", "add three", add_three)
        plugin_mgr.register(add_three_plugin)

        def division(ctx, a, b):
            return a / b
        division_plugin = Plugin.wrap_function(
            "division", "division", division)
        plugin_mgr.register(division_plugin)

        def add_four(ctx, num):
            if num < 0:
                raise InvalidArgumentException("num must more than zero!")
            return num + 4
        add_four_plugin = Plugin.wrap_function(
            "add_four", "add four", add_four)
        plugin_mgr.register(add_four_plugin)

        plugin_mgr.sys_prepare(None)
        self.plugin_mgr = plugin_mgr

    def test_execute_common(self):
        # 测试普通运行时的情况
        worklist = (
            Job("add_one"),
            Job("add_three", args=("$add_one.result",)),
            Decision("judge",
                     lambda ctx: "test_end"
                     if ctx["add_three"] > 5 else "add_two"),
            Job("add_two", args={"num": "$add_three"}, goto="end"),
            OkEnd("test_end", execute=lambda ctx: ctx["add_three"])  # 自定义end
        )

        workflow = Workflow(worklist, None, self.plugin_mgr)
        end = workflow.execute(args={"add_one": (1,)})
        self.assertEquals(end.result, 7)

        end = workflow.execute(args={"add_one": {"num": 10}})
        self.assertEquals(end.result, 14)

    def test_execute_with_exception(self):
        # 测试异常发生时的情况
        worklist = (
            Job("add_one", args=(5, )),
            Job("division", args={"a": "$add_one.result"})
        )

        workflow = Workflow(worklist, None, self.plugin_mgr)
        end = workflow.execute(args={
            "division": {"b": 3}
        })
        self.assertEquals(end.result, 2)

        # 除0错误
        end = workflow.execute(args={
            "division": {"b": 0}
        })
        self.assertEquals(end.status, End.STATUS_ERROR_HAPPENED)
        self.assertEquals(end.exc_type, ZeroDivisionError)

    def test_execute_with_invalid_args(self):
        # 测试错误请求参数
        worklist = (
            Job("add_four"),
        )

        workflow = Workflow(worklist, None, self.plugin_mgr)
        end = workflow.execute(args={"add_four": (1, )})
        self.assertEquals(end.result, 5)

        end = workflow.execute(args={"add_four": (-1, )})
        self.assertEquals(end.status, End.STATUS_BAD_REQUEST)

    def test_add_listener(self):
        listener = ListenerA(0)
        worklist = (
            Job("add_one"),
            Job("add_two", args=("$add_one.result",)),
            Decision("decide", lambda ctx: "add_three"
                     if ctx["add_two.result"] > 10 else "division"),
            Job("add_three", args=("$add_two.result",)),
            OkEnd("add_three_end", execute=lambda ctx: ctx["add_three"]),
            Job("division", args=("$add_two.result", 2))
        )

        workflow = Workflow(worklist, None, self.plugin_mgr)
        workflow.add_listener(listener)
        workflow.add_listener(ListenerB)

        end = workflow.execute(args={"add_one": (1, )})
        self.assertEquals(end.result, 2)
        self.assertEquals(listener.a, -4)

        listener.a = 0
        end = workflow.execute(args={"add_one": (10, )})
        self.assertEquals(end.result, 16)
        self.assertEquals(listener.a, -5)

        listener.a = 0
        end = workflow.execute(args={
            "add_one": (1, ),
            "division": ("$add_two.result", 0)
        })
        self.assertEquals(end.status, End.STATUS_ERROR_HAPPENED)
        self.assertEquals(listener.a, -2)
        self.assertEquals(end.exc_type, ZeroDivisionError)


class ListenerA(AbstractListener):

    def __init__(self, a):
        self.a = a

    def on_start(self, ctx):
        print "\n> listener a on workflow start"
        ctx["a"] = self.a

    def on_unit_start(self, ctx):
        print "> listener a enter unit {}".format(ctx.current_unit)
        self.a += 1

    def on_unit_finish(self, ctx):
        print "> listener a finish unit {}".format(ctx.current_unit)
        self.a -= 2

    def on_finish(self, ctx):
        print "> listener a on workflow finish"
        ctx["a_finish"] = self.a + 1

    def on_error(self, ctx, exc_type, exc_value, tb):
        print "> error happened on unit {}".format(ctx.current_unit)
        self.exc_type = exc_type


class ListenerB(AbstractListener):

    def on_start(self, ctx):
        self.x = 0

    def on_unit_start(self, ctx):
        self.x += 1

    def on_finish(self, ctx):
        print "\n In listener B x is {}".format(self.x)



# coding: utf-8

import os
import os.path
import sqlite3
from girlfriend.plugin.orm import (
    EngineManager,
    _engine_manager,
    Query,
    SQL,
    KeyExtractWrapper
)
from girlfriend.util.config import Config
from girlfriend.exception import (
    InvalidArgumentException,
    InvalidStatusException
)
from girlfriend.testing import GirlFriendTestCase
from sqlalchemy import (
    Column,
    Integer,
    String,
    func,
    and_
)
from sqlalchemy.pool import QueuePool
from sqlalchemy.ext.declarative import declarative_base
from girlfriend.plugin import plugin_manager
from girlfriend.data.table import TableWrapper, Title


class EngineManagerTestCase(GirlFriendTestCase):

    def setUp(self):
        self.test_db_file = "{}/gftest.db".format(os.environ["HOME"])
        connection_url = "sqlite:///{}".format(self.test_db_file)
        self.config = Config({
            "db_test": {
                "connect_url": "sqlite:///:memory:",
                "pool_policy": "test"
            },
            "db_test2": {
                "connect_url": connection_url
            },
            "dbpool_test": {
                "poolclass": "QueuePool",
                "pool_size": 10,
                "pool_recycle": 3600,
                "pool_timeout": 20
            }
        })
        conn = sqlite3.connect(self.test_db_file)
        create_table_sql = """
        create table user (
            id integer primary key,
            name varchar(10) unique,
            grade int not null,
            description text not null
        )
        """
        conn.execute(create_table_sql)
        conn.execute((
            "insert into user (id, name, grade, description) values "
            "(1, 'SamChi', 1, 'I am SamChi')"
        ))
        conn.commit()

    def tearDown(self):
        if os.path.exists(self.test_db_file):
            os.remove(self.test_db_file)

    def test_validate(self):
        engine_manager = EngineManager()
        engine_manager.validate_config(self.config)

        # 是否已经设置默认参数
        self.assertEquals(self.config["db_test"]["connect_args"], "{}")
        self.assertEquals(self.config["db_test"]["encoding"], "utf-8")

        bad_config = Config({
            "db_test": {
                "connect_url": "hehe"  # url format error
            }
        })

        # 因为已经验证过配置了，所以同一个manager对象不会进行重复验证
        engine_manager.validate_config(bad_config)  # No Exception happened

        engine_manager = EngineManager()
        self.failUnlessException(
            InvalidArgumentException,
            engine_manager.validate_config, bad_config)

        bad_config["db_test"]["connect_url"] = "sqlite:///:memory:"
        bad_config["db_test"]["connect_args"] = "{h"
        engine_manager = EngineManager()
        self.failUnlessException(
            InvalidArgumentException,
            engine_manager.validate_config, bad_config)

    def test_init_all(self):
        engine_manager = EngineManager()
        # 同一个manager对象，重复初始化不会发生任何问题
        for i in xrange(0, 100):
            engine_manager.validate_config(self.config)
            engine_manager.init_all(self.config)

            test_engine = engine_manager.engine("test")
            self.assertIsNotNone(test_engine)

            # 连接池相关参数
            pool = test_engine.engine.pool
            self.assertIsInstance(pool, QueuePool)
            self.assertEquals(pool.size(), 10)
            self.assertEquals(pool._recycle, 3600)
            self.assertEquals(pool._timeout, 20)

            # 测试可否连接
            connection = test_engine.engine.connect()
            result = connection.execute("select 1 + 1")
            self.assertEquals(tuple(result)[0][0], 2)

            test_engine2 = engine_manager.engine("test2")
            self.assertIsNotNone(test_engine2)

            connection = test_engine2.engine.connect()
            result = connection.execute("select * from user")
            self.assertEquals(tuple(result)[0],
                              (1, "SamChi", 1, "I am SamChi"))

    def test_dispose_all(self):
        engine_manager = EngineManager()
        engine_manager.validate_config(self.config)
        engine_manager.init_all(self.config)
        engine_manager.dispose_all()

        self.failUnlessException(InvalidStatusException,
                                 engine_manager.engine, "test")

Base = declarative_base()


class Student(Base):

    __tablename__ = "student"

    id = Column(Integer, primary_key=True)

    name = Column(String(10))

    fullname = Column(String(20))

    grade = Column(Integer)


class Grade(Base):

    __tablename__ = "grade"

    id = Column(Integer, primary_key=True)

    name = Column(String(10))


class QueryTestCase(GirlFriendTestCase):

    def setUp(self):
        config = Config({
            "db_test": {
                "connect_url": "sqlite:///:memory:",
                "pool_policy": "test"
            },
            "dbpool_test": {
                "poolclass": "QueuePool",
                "pool_size": 10,
            }
        })
        _engine_manager.validate_config(config)
        _engine_manager.init_all(config)
        global Base
        engine_container = _engine_manager.engine("test")
        Base.metadata.create_all(engine_container.engine)

        student_sam = Student(id=1, name="Sam", fullname="SamChi", grade=1)
        student_jack = Student(id=2, name="Jack", fullname="JackMa", grade=2)
        student_betty = Student(
            id=3, name="Betty", fullname="Betty Smith", grade=1)
        grade1 = Grade(id=1, name="Grade One")
        grade2 = Grade(id=2, name="Grade Two")

        session = engine_container.session()
        session.add_all([
            student_sam,
            student_jack,
            student_betty,
            grade1,
            grade2
        ])
        session.commit()
        self.config = config

    def test_query(self):
        ctx = {}

        # 执行纯SQL语句的情景
        q = Query(
            "test", "test_result", Student,
            "select * from student order by id")
        q(ctx)
        self._check_students_list(ctx["test_result"])

        # 进行auto map的情况
        q = Query("test", "test_result", "student",
                  "select * from student order by id")
        q(ctx)
        self._check_students_list(ctx["test_result"])

        # 多个映射类的情况
        q = Query("test", "test_result",
                  (Student.id, Student.name, Grade.name),
                  ("select s.id, s.name, g.name from student s, grade g "
                   "where s.grade = g.id order by s.id"))
        q(ctx)
        self.assertTrue(ctx["test_result"][0], (1, "Sam", "Grade One"))

        # auto map field 的情况
        q = Query(
            "test", "test_result",
            (Student.id, "student.name", "grade.name"),
            ("select s.id, s.name, g.name from student s, grade g "
             "where s.grade = g.id order by s.id"))
        q(ctx)
        self.assertTrue(ctx["test_result"][0], (1, "Sam", "Grade One"))

        # 查询字符串表达式
        q = Query("test", "test_result", Student,
                  "id > :id and name = :name",
                  params={"id": 1, "name": "Betty"})
        q(ctx)
        students = ctx["test_result"]
        self.assertEquals(len(students), 1)
        s = students[0]
        self.assertEquals((s.id, s.name, s.grade), (3, "Betty", 1))

        # order by
        q = Query("test", "test_result", Student,
                  "id > :id", order_by="id desc", params={"id": 1})
        q(ctx)
        self.assertEquals([std.id for std in ctx["test_result"]], [3, 2])

        # group by and count
        q = Query("test", "test_result",
                  ("student.grade", func.count('*')), order_by="grade desc",
                  group_by="student.grade")
        q(ctx)
        grade_counts = ctx["test_result"]
        self.assertEquals(len(grade_counts), 2)
        self.assertEquals(grade_counts[0], (2, 1))

        # 条件查询
        q = Query("test", "test_result", Student, and_(
            Student.grade == 1, Student.name == "Betty"))
        q(ctx)
        records = ctx["test_result"]
        self.assertEquals(len(records), 1)

        # 通过自定义函数来构建查询条件
        def my_query(session, ctx, student, grade):
            query = session.query(student.id, grade.name).filter(
                student.grade == grade.id)
            return query.all()

        #  只返回查询对象
        q_a = Query("test", "test_result", ("student", "grade"),
                    lambda s, ctx, std, grade:
                    s.query(std.id, grade.name).filter(std.grade == grade.id))
        #  通过all()返回查询结果
        q_b = Query("test", "test_result", ("student", "grade"),
                    lambda s, ctx, std, grade:
                    s.query(std.id, grade.name).filter(
                        std.grade == grade.id).all())
        for q in (q_a, q_b):
            q(ctx)
            records = ctx["test_result"]
            self.assertEquals(len(records), 3)
            self.assertEquals(records[0], (1, "Grade One"))

        class StudentGrade(object):

            def __init__(self, row):
                student, grade = row
                self.id = student.id
                self.name = student.name
                self.grade = grade.name

        # row_handler
        q = Query("test", "test_result", (Student, Grade),
                  Student.grade == Grade.id, row_handler=StudentGrade,
                  order_by=Student.id)
        q(ctx)
        records = ctx["test_result"]
        self.assertEquals(len(records), 3)
        self.assertEquals(
            (records[0].id, records[0].name, records[0].grade),
            (1, "Sam", "Grade One")
        )

        # Table Wrapper
        q = Query(
            "test", "test_table",
            (Student.id, Student.name, Student.grade),
            result_wrapper=TableWrapper(
                "student",
                (
                    Title("id", u"编号"),
                    Title("name", u"姓名"),
                    Title("grade", u"年级")
                ))
        )
        q(ctx)
        test_table = ctx["test_table"]
        self.assertEquals(test_table.cell(0, 0), 1)
        self.assertEquals(tuple(test_table.row(0)), (1, u"Sam", 1))

        # Pure SQL Query
        s = SQL("test", "sql_1", "select * from student order by id")
        s(ctx)
        records = ctx["sql_1"]
        self.assertEquals(records[0], (1, "Sam", "SamChi", 1))

        def my_row_handler(row):
            return Student(id=row[0], name=row[1],
                           fullname=row[2], grade=row[3])

        s = SQL("test", "sql_2", "select * from student where id > :id",
                {"id": 1}, row_handler=my_row_handler)
        s(ctx)
        records = ctx["sql_2"]
        self.assertEquals((records[0].id, records[0].name), (2, "Jack"))

        # Table Wrapper
        s = SQL("test", "sql_3", "select * from student order by id",
                result_wrapper=TableWrapper("students",
                                            (
                                                Title("id", u"编号"),
                                                Title("name", u"名字"),
                                                Title("fullname", u"全名"),
                                                Title("grade", u"年级")
                                            ))
                )
        s(ctx)
        table = ctx["sql_3"]
        row = table[0]
        self.assertEquals((row.id, row.name), (1, "Sam"))

        # 测试非查询语句
        s = SQL("test", None,
                (
                    "insert into student (id, name, fullname, grade) "
                    "values (:id, :name, :fullname, :grade)"
                ),
                {"id": 4, "name": "Jane", "fullname": "JaneGreen", "grade": 3})
        s(ctx)
        SQL("test", "sql_4", "select * from student where id = 4")(ctx)
        self.assertEquals(ctx["sql_4"][0], (4, "Jane", "JaneGreen", 3))

        # Test Plugin
        plugin_manager.sys_prepare(self.config)
        orm_query = plugin_manager.plugin("orm_query")
        orm_query.execute(
            ctx,
            Query("test", "result_1", (Student.id, Student.name), "id = 1"),
            Query("test", "result_2", (Student.id, Student.name), "id = 2")
        )
        result_1, result_2 = ctx["result_1"], ctx["result_2"]
        self.assertEquals([(1, "Sam")], result_1)
        self.assertEquals([(2, "Jack")], result_2)

    def _check_students_list(self, students):
        self.assertEquals(len(students), 3)
        self.assertEquals([std.name for std in students],
                          ["Sam", "Jack", "Betty"])


class KeyExtratcWrapperTestCase(GirlFriendTestCase):

    def test_wrap_collection_row(self):
        table = (
            (1, "Sam", 26, "A"),
            (2, "Jack", 25, "A"),
            (3, "Peter", 26, "B")
        )

        result = KeyExtractWrapper(0)(table)
        self.assertEquals(
            result,
            {
                1: (1, "Sam", 26, "A"),
                2: (2, "Jack", 25, "A"),
                3: (3, "Peter", 26, "B")
            }
        )

        table = (
            {"id": 1, "name": "Sam", "grade": "A"},
            {"id": 2, "name": "Jack", "grade": "B"}
        )

        result = KeyExtractWrapper("name")(table)
        self.assertEquals(
            result,
            {
                "Sam": {"id": 1, "name": "Sam", "grade": "A"},
                "Jack": {"id": 2, "name": "Jack", "grade": "B"}
            }
        )

    def test_wrap_object(self):
        table = (
            Grade(id=1, name="A"),
            Grade(id=2, name="B")
        )

        result = KeyExtractWrapper("id")(table)
        self.assertEquals(
            result,
            {
                1: table[0],
                2: table[1]
            }
        )



# coding: utf-8

"""lang.validating测试用例
"""

from numbers import Number
from girlfriend.util.validating import Rule
from girlfriend.exception import InvalidArgumentException
from girlfriend.testing import GirlFriendTestCase


class RuleTestCase(GirlFriendTestCase):

    def test_validate(self):

        def test_logic(value):
            if not isinstance(value, Number):
                return
            if value % 2 != 0:
                return u"{} 不能被2整除".format(value)

        test_rule = Rule(
            "test",
            type=(str, int),
            required=True,
            min=5,
            max=20,
            regex=r"^\d+$",
            logic=test_logic
        )

        # 正确通过验证的情况
        test_rule.validate("1" * 6)
        test_rule.validate(20)

        # 错误的类型
        self.failUnlessException(
            InvalidArgumentException, test_rule.validate, 1.1)

        # 错误的长度
        self.failUnlessException(
            InvalidArgumentException, test_rule.validate, "hehe")
        self.failUnlessException(
            InvalidArgumentException, test_rule.validate, 2)
        self.failUnlessException(
            InvalidArgumentException, test_rule.validate, "h" * 30)
        self.failUnlessException(
            InvalidArgumentException, test_rule.validate, 30)

        # 正则验证
        self.failUnlessException(
            InvalidArgumentException, test_rule.validate, "1a2b3c")

        # 业务逻辑验证
        self.failUnlessException(
            InvalidArgumentException, test_rule.validate, 11)



# coding: utf-8

"""util.lang模块的测试用例
"""

from girlfriend.testing import GirlFriendTestCase
from girlfriend.util.lang import args2fields


class Args2FiledsTestCase(GirlFriendTestCase):

    def test_args2fileds(self):

        class A(object):

            @args2fields()
            def __init__(self, a, b=2, c=5):
                self.d = self._a + self._b + self._c

        a = A(1, 2, 3)
        self.assertEquals(a.d, 6)

        a = A(a=1, b=3)
        self.assertEquals(a.d, 9)

        a = A(1)
        self.assertEquals(a.d, 8)



# coding: utf-8

"""
该工作流可以快速将SQL导出为Excel表格，并通过邮件发送给目标用户
使用json格式的文件来描述SQL任务，例如：

# 该单元为SQL任务，db为连接的数据库，sql为要执行的SQL语句，table为表名称，同时会用作Excel Sheet名称
# titles 为表格各字段的名称，必须按照顺序
{
  "db": "user",
  "sql": "select id, email, cat_num from user",
  "table": "注册用户",
  "titles": ["编号", "姓名", "宠物猫数目"]
}
{
  "db": "product",
  "sql": "select id, name, sell_num from product",
  "table": "商品销售详情",
  "titles": ["编号", "姓名", "销售数目"]
}

# 该单元描述邮件发送详情，可选，server为要使用的smtp服务器，sender为发件人，receivers为收件人
  当receivers为逗号隔开的字符串时，会一并发送给所有的收件人，方便统一沟通，当receivers为数组时，会单独发给每个人
  content为html形式的邮件正文，table为要在正文显示的html表格
{
  "server": "test",
  "sender": "hongze.chi@gmail.com",
  "receivers": "chihz3800@163.com",
  "subject": "数据邮件",
  "content": "<h1>数据报表</h1>",
  "table": [0, 1],
}

# 该单元描述导出Excel文件，sheets为包含的数据库表，从0开始计数
{
  "workbook": "test.xlsx",
  "sheets": [0, 1]
}

使用方法：
    gf_workflow -m :sqlreport -t sqltask.json
"""

import argparse
import os.path
from girlfriend.data.table import TableWrapper
from girlfriend.util.script import show_msg_and_exit
from girlfriend.workflow.gfworkflow import Job, Decision
from girlfriend.plugin.json import JSONR
from girlfriend.plugin.orm import SQL
from girlfriend.plugin.excel import SheetW
from girlfriend.plugin.table import HTMLTable
from girlfriend.plugin.mail import Attachment


cmd_parser = argparse.ArgumentParser(description=__doc__.decode("utf-8"))
cmd_parser.add_argument("--task", "-t", dest="task", default="",
                        help=u"指定任务描述文件")
cmd_parser.add_argument("--print", "-p", dest="print_tables",
                        action="store_true", help=u"是否打印表格")


def workflow(options):
    # 检查配置文件
    if not options.task:
        show_msg_and_exit(u"必须指定一个任务描述文件，文件格式见帮助")
    if not os.path.exists(options.task):
        show_msg_and_exit(u"任务描述文件 '{}' 不存在".format(options.task))

    return (
        # 读取json格式的任务描述文件并保存到上下文
        Job(
            "read_json",
            args=(
                JSONR(options.task, "block",
                      result_wrapper=_task, variable="task"),
            )
        ),

        # 执行SQL语句
        Job(
            "orm_query",
            args=_gen_orm_query_args
        ),

        # 决定是否打印表格到终端
        Decision(
            "need_print",
            lambda ctx: "print_table"
            if options.print_tables else "need_write_excel"
        ),

        # print tables
        Job(
            "print_table",
            args="orm_query.result"
        ),

        # 决定是否要输出Excel
        Decision(
            "need_write_excel",
            lambda ctx: "write_excel"
            if ctx["task"].get("workbooks") else "need_send_mail",
        ),

        # 输出Excel
        Job(
            "write_excel",
            args=_gen_excel_args
        ),

        # 是否发送Email
        Decision(
            "need_send_mail",
            lambda ctx: "need_render_html"
            if ctx["task"].get("mail") else "end"
        ),

        # 是否需要渲染html表格
        Decision(
            "need_render_html",
            lambda ctx: "html_table"
            if ctx["task"]["mail"].get("tables") else "send_mail"
        ),

        # 渲染html表格
        Job(
            "html_table",
            args=lambda ctx: [HTMLTable(
                ctx["orm_query.result"][idx],
                property=ctx["task"]["mail"].get("table_property"))
                for idx in ctx["task"]["mail"]["tables"]]
        ),

        # 发送邮件了!
        Job(
            "send_mail",
            args=_gen_send_mail_args
        )
    )


def _task(records):
    task = {"sqltasks": [], "workbooks": []}
    for record in records:
        if "db" in record:
            task["sqltasks"].append(record)
        elif "server" in record:
            task["mail"] = record
        elif "workbook" in record:
            task["workbooks"].append(record)
    return task


def _gen_orm_query_args(ctx):
    sqltasks = ctx["task"]["sqltasks"]
    return [SQL(
        engine_name=t["db"],
        variable_name="table_%d" % idx,
        sql=t["sql"],
        result_wrapper=TableWrapper(t["table"], t["titles"],
                                    auto_title_name=True)
    )for idx, t in enumerate(sqltasks)]


def _gen_excel_args(ctx):
    tables = ctx["orm_query.result"]
    for idx, workbook in enumerate(ctx["task"]["workbooks"]):
        yield {
            "filepath": workbook["workbook"],
            "sheets": [SheetW(tables[i]) for i in workbook["sheets"]]
        }


def _gen_send_mail_args(ctx):
    args = {}
    args.update(ctx["task"]["mail"])
    tables = args.pop("tables", None)
    args.pop("table_property", None)
    if tables:
        content = args["content"]
        for html_table in ctx["html_table.result"]:
            content += (html_table + "<br/>")
        args["content"] = content
    args["attachments"] = [
        Attachment(w["workbook"], "application/octet-stream")
        for w in ctx["task"]["workbooks"]
    ]
    return args



# coding: utf-8

"""json plugin testcase
"""

import os
import ujson
import httpretty

from girlfriend.testing import GirlFriendTestCase
from girlfriend.data.table import (
    TableWrapper,
    Title
)
from girlfriend.plugin.json import JSONR, JSONW

JSON_LINE_CONTENT = """
{"id": 1, "name": "Sam", "grade": "A"}
{"id": 2, "name": "Tom", "grade": "B"}
{"id": 3, "name": "Betty", "grade": "A"}
"""

JSON_ARRAY_CONTENT = """
[
  {"id": 1, "name": "Jack", "grade": "A"},
  {"id": 2, "name": "Peter", "grade": "B"},
  {"id": 3, "name": "James", "grade": "A"}
]
"""

JSON_OBJECT_CONTENT = """
{
  "code": 200,
  "value": {
    "users": [
      {"id": 1, "name": "Sam", "grade": "A"},
      {"id": 2, "name": "Jack", "grade": "A"},
      {"id": 3, "name": "James", "grade": "A"}
    ]
  }
}
"""

JSON_BLOCK_CONTENT = """
hhhhhhhhsdddsddwwdsdd
{
    "name": "Sam",
    "scores": {
        "A": 102,
        "B": 110,
        "C": 115
    }
}sssssssss{
    "name": "Jack",
    "scores": {
        "A": 110,
        "B": 50,
        "C": 60,
        "D": "{-_-}"
    }
}
"""


class Student(object):

    def __init__(self, id, name, grade):
        self.id = id
        self.name = name
        self.grade = grade


class JSONReaderPluginTestCase(GirlFriendTestCase):

    def setUp(self):
        with open("line_file.json", "w") as f:
            f.write(JSON_LINE_CONTENT)
        with open("array_file.json", "w") as f:
            f.write(JSON_ARRAY_CONTENT)
        with open("object_file.json", "w") as f:
            f.write(JSON_OBJECT_CONTENT)
        with open("block_file.json", "w") as f:
            f.write(JSON_BLOCK_CONTENT)

        httpretty.enable()

        httpretty.register_uri(
            httpretty.GET,
            "http://test.gf/line",
            body=JSON_LINE_CONTENT,
            content_type="text/plain"
        )

        httpretty.register_uri(
            httpretty.GET,
            "http://test.gf/array",
            body=JSON_ARRAY_CONTENT,
            content_type="application/json"
        )

        httpretty.register_uri(
            httpretty.GET,
            "http://test.gf/object",
            body=JSON_OBJECT_CONTENT,
            content_type="application/json"
        )

    def test_read_line_with_file(self):
        ctx = {}

        # 测试一般执行
        result = JSONR("line_file.json", "line")(ctx)
        self.assertEquals(len(result), 3)
        self.assertEquals(result[0], {"id": 1, "name": "Sam", "grade": "A"})

        # 测试record_handler
        JSONR("line_file.json", "line",
              record_handler=lambda row: {row.pop("name"): row},
              variable="test")(ctx)
        test_result = ctx["test"]
        self.assertEquals(test_result[0], {"Sam": {"id": 1, "grade": "A"}})

        # 测试result_wrapper
        table = JSONR("line_file.json", "line",
                      result_wrapper=TableWrapper("students", (
                          Title("id", u"编号"),
                          Title("name", u"姓名"),
                          Title("grade", u"等级")
                      )))(ctx)
        print "\n", table
        row_0 = table[0]
        self.assertEquals(
            (row_0.id, row_0.name, row_0.grade),
            (1, "Sam", "A")
        )

    def test_read_line_with_web(self):
        ctx = {}

        # 测试一般执行
        result = JSONR("http://test.gf/line", "line")(ctx)
        self.assertEquals(len(result), 3)
        self.assertEquals(result[0], {"id": 1, "name": "Sam", "grade": "A"})

    def test_read_json_array(self):
        ctx = {}

        # 测试一般执行
        result = JSONR("array_file.json", "array")(ctx)
        self.assertEquals(len(result), 3)
        self.assertEquals(result[0], {"id": 1, "name": "Jack", "grade": "A"})

        # URL加载执行
        result = JSONR("http://test.gf/array", "array")(ctx)
        self.assertEquals(len(result), 3)
        self.assertEquals(result[0], {"id": 1, "name": "Jack", "grade": "A"})

        # 测试record_handler
        result = JSONR("http://test.gf/array", "array",
                       record_handler=lambda r: Student(**r))(ctx)
        std = result[0]
        self.assertEquals((std.id, std.name, std.grade), (1, "Jack", "A"))

        table = JSONR("http://test.gf/array", "array",
                      record_handler=lambda r: Student(**r),
                      result_wrapper=TableWrapper(
                          "student",
                          (Title("id"), Title("name"), Title("grade"))))(ctx)
        print "\n", table
        row_0 = table[0]
        self.assertEquals(
            (row_0.id, row_0.name, row_0.grade),
            (1, "Jack", "A")
        )

    def test_read_json_object(self):
        ctx = {}

        # 测试从文件提取
        result = JSONR("object_file.json", "extract:value.users")(ctx)
        self.assertEquals(len(result), 3)
        self.assertEquals(result[0], {"id": 1, "name": "Sam", "grade": "A"})

        # 测试从web提取
        result = JSONR("http://test.gf/object", "extract:value.users")(ctx)
        self.assertEquals(len(result), 3)
        self.assertEquals(result[0], {"id": 1, "name": "Sam", "grade": "A"})

        # 测试record_handler
        result = JSONR("http://test.gf/object", "extract:value.users",
                       record_handler=lambda r: Student(**r))(ctx)
        self.assertEquals(len(result), 3)
        std = result[0]
        self.assertEquals((std.id, std.name, std.grade), (1, "Sam", "A"))

        # 测试result_wrapper
        table = JSONR("http://test.gf/object", "extract:value.users",
                      record_handler=lambda r: Student(**r))(ctx)
        self.assertEquals(len(result), 3)
        row_0 = table[0]
        self.assertEquals(
            (row_0.id, row_0.name, row_0.grade),
            (1, "Sam", "A")
        )

    def test_read_json_blocks(self):
        ctx = {}

        result = JSONR("block_file.json", "block")(ctx)
        self.assertEquals(len(result), 2)
        self.assertEquals(result[0], {"name": "Sam",
                                      "scores": {
                                          "A": 102,
                                          "B": 110,
                                          "C": 115
                                      }})

    def tearDown(self):
        os.remove("line_file.json")
        os.remove("array_file.json")
        os.remove("object_file.json")
        os.remove("block_file.json")
        httpretty.disable()
        httpretty.reset()


class JSONWriterPluginTestCase(GirlFriendTestCase):

    def setUp(self):
        self.records = [
            {"id": 1, "name": "Sam", "gender": 1},
            {"id": 2, "name": "Jack", "gender": 1},
            {"id": 3, "name": "Betty", "gender": 0},
        ]
        httpretty.enable()

    def test_write_json_file_by_line(self):
        JSONW("line_file.json", "line", self.records)({})
        with open("line_file.json", "r") as f:
            for idx, line in enumerate(f):
                json_obj = ujson.loads(line.strip())
                self.assertEquals(json_obj, self.records[idx])
        os.remove("line_file.json")

    def test_write_object(self):
        JSONW("object_file.json", "object", self.records[0])({})
        with open("object_file.json", "r") as f:
            self.assertEquals(ujson.loads(f.read()), self.records[0])
        os.remove("object_file.json")

    def test_write_array(self):
        JSONW("array_file.json", "array", self.records)({})
        with open("array_file.json", "r") as f:
            self.assertEquals(ujson.loads(f.read()), self.records)
        os.remove("array_file.json")

    def tearDown(self):
        httpretty.disable()
        httpretty.reset()



# coding: utf-8

"""本模块提供命令行参数解析辅助工具
"""

from termcolor import colored
import argparse


def print_help(parser, color=None):
    """输出帮助信息"""
    help_text = []
    for action in parser._actions:
        if isinstance(action, argparse._HelpAction):
            continue
        option_strings = ",".join(action.option_strings)
        help_text.append(u"{}  {}".format(option_strings, action.help))
    print "\n", colored("\n".join(help_text), color)



# coding: utf-8

"""所有Girlfriend内部公共异常都定义在这里
"""

import types
import locale


class GirlFriendException(Exception):

    """GirlFriend系统中所有的异常都要继承此异常类
    """

    def __init__(self, msg):
        """
            :param msg 可以是一个字符串，也可以是一个key为locale的字典
        """
        if isinstance(msg, types.DictType):
            country_code, _ = locale.getlocale(locale.LC_ALL)
            msg = msg[country_code]

        if isinstance(msg, unicode):
            super(GirlFriendException, self).__init__(msg.encode("utf-8"))
            self.msg = msg
        elif isinstance(msg, str):
            super(GirlFriendException, self).__init__(msg)
            self.msg = msg.decode("utf-8")
        else:
            raise TypeError

    def __unicode__(self):
        return self.msg


class GirlFriendBizException(GirlFriendException):

    """业务相关的异常，属于意料之中的用户错误，捕获该异常之后可以按照正常的业务流程进行处理，
       常用于面向最终用户的Facade层。最常见的场景就是在命令行工具中，可以捕获该异常，
       向用户输出错误消息并退出或者重新输入参数，因为是正常的逻辑错误，无需上报Bug和输出错误日志。
    """
    pass


class GirlFriendSysException(GirlFriendException):

    """系统相关的异常，多用于框架内部以及向插件开发者提供的工具和数据结构，
       该异常跟业务无关，属于系统错误，无法继续进行业务处理，
       只能输出错误日志、打印错误堆栈进行调试或者将错误上报给开发者。
    """
    pass


class InvalidArgumentException(GirlFriendSysException):

    """当函数输入的参数不正确时抛出此异常
    """
    pass


class InvalidTypeException(GirlFriendSysException):

    """当接受的参数对象类型不符合预期时抛出此异常
    """
    pass


class UnknownWitchToExecuteException(GirlFriendSysException):

    """多数用于委托的场景，如果委托的对象有多种表现形式，指定的表现形式冲突，
       就会抛出此异常，比如一个Job既可以接受插件，又可以接受普通的回调函数，如果
       两个都指定了，那么就会不知道到底该执行哪个
    """
    pass


class UnsupportMethodException(GirlFriendSysException):

    """当实现一个抽象类时，如果某些方法无法满足约定实现，那么可以抛出该异常，
       外部调用该对象的框架必须知道并处理这种异常情况。
    """
    pass


class InvalidStatusException(GirlFriendSysException):

    """当状态不满足时，抛出此异常
    """
    pass



# coding: utf-8

"""本模块规范了workflow各个组件所需满足的接口形式，
在girlfriend中，一个完整的workflow由以下组件构成:

   Start       任务执行的起始点，列表中的第一个Job默认为起始点

   Job         任务执行的基本逻辑单元

   Decision    决策单元，当工作流有多个逻辑分支时
               对上一步的执行结果进行判断，决定接下来运行的Job

   WaitingFor  等待单元,等待某个条件到达再继续执行接下来的工作,
               比如某个时间点,某个队列消息,某个文件变更等等

   End         任务结束点，携带执行状态和最终返回结果

   Context     上下文对象，用于在工作流的不同Job之间传递数据

   Workflow    工作流本身，包含要执行的工作流描述以及具体的执行逻辑

   Listener    分为针对单个Job的事件监听和针对整个workflow的工作监听
"""

import traceback
import collections
import itertools
from abc import (
    ABCMeta,
    abstractmethod,
    abstractproperty
)
from girlfriend.exception import InvalidArgumentException
from girlfriend.util.lang import args2fields
from girlfriend.util.config import Config


class WorkflowUnit(object):

    """工作流单元抽象
       Start Job Decision WaitFor End隶属于工作流单元
    """
    __metaclass__ = ABCMeta

    @abstractproperty
    def name(self):
        """workflow中的每一个单元都要有一个独一无二的名字
        """
        pass

    @abstractproperty
    def unittype(self):
        """工作单元类型
        """
        pass

    @abstractmethod
    def execute(self, context):
        """执行该单元的具体逻辑
           该方法由执行引擎进行回调
        """
        pass


class AbstractJob(WorkflowUnit):

    """任务执行单元抽象
    """

    __metaclass__ = ABCMeta

    @property
    def unittype(self):
        return "job"

    @abstractproperty
    def goto(self):
        pass


class AbstractDecision(WorkflowUnit):

    """决策单元抽象
    """

    __metaclass__ = ABCMeta

    @property
    def unittype(self):
        return "decision"


class AbstractWaitingFor(WorkflowUnit):

    """等待单元
    """

    __metaclass__ = ABCMeta

    @property
    def unittype(self):
        return "waiting_for"


class End(WorkflowUnit):

    """结束单元
    """

    STATUS_OK = 0  # 一切顺利,任务成功结束

    STATUS_BAD_REQUEST = 1  # 请求参数错误,任务无法执行

    STATUS_ERROR_HAPPENED = 2  # 服务出错

    def __init__(self, name, status, result=None, execute=None):
        self._name = name
        self._status = status
        self._result = result
        self._execute = execute

    @property
    def name(self):
        return self._name

    @property
    def unittype(self):
        return "end"

    @property
    def status(self):
        """状态码,描述本次工作流的执行结果
        """
        return self._status

    @property
    def result(self):
        """执行结果,工作流执行的最终汇总结果
        """
        return self._result

    def execute(self, context):
        if self._execute:
            self._result = self._execute(context)

    def __str__(self):
        return "<End name={name}, status={status}, result={result}>".format(
            name=self.name,
            status=self.status,
            result=self.result
        )


class OkEnd(End):

    """顺利执行时返回的结果
    """

    def __init__(self, name="end", result=None, execute=None):
        End.__init__(
            self,
            name=name,
            status=End.STATUS_OK,
            result=result,
            execute=execute
        )


class BadRequestEnd(End):

    """请求参数错误时的返回结果
    """

    def __init__(self, name="bad_request_end",
                 msg=None, result=None, execute=None):
        End.__init__(
            self,
            name=name,
            status=End.STATUS_BAD_REQUEST,
            result=result,
            execute=execute
        )
        self._msg = msg

    @property
    def msg(self):
        return self._msg


class ErrorEnd(End):

    """当出现程序异常时返回的结果
    """

    def __init__(self, exc_type, exc_value, tb,
                 name="error_end", result=None, execute=None):
        """
            :param exc_type 异常类型
            :param exc_value 异常值
            :param tb 异常发生的traceback
        """
        End.__init__(
            self,
            name=name,
            status=End.STATUS_ERROR_HAPPENED,
            result=result,
            execute=execute
        )
        self._exc_type = exc_type
        self._exc_value = exc_value
        self._tb = tb

    @property
    def exc_type(self):
        return self._exc_type

    @property
    def exc_value(self):
        return self._exc_value

    @property
    def tb(self):
        return self._tb

    def print_exc(self):
        """打印异常
        """
        traceback.print_exception(self._exc_type, self._exc_value, self._tb)


class AbstractContext(collections.Mapping):

    """上下文抽象
       上下文由执行引擎在每次执行之前选择具体类型进行创建
    """

    __metaclass__ = ABCMeta

    def __init__(self):
        super(AbstractContext, self).__init__()

    @abstractproperty
    def config(self):
        """全局配置
        """
        pass

    @abstractproperty
    def logger(self):
        """获取日志对象
        """
        pass

    @abstractproperty
    def current_unit(self):
        """当前工作流正执行的单元名称
        """
        pass

    @abstractproperty
    def current_unittype(self):
        """当前执行单元类型
        """
        pass

    @abstractmethod
    def args(self, job_name):
        """获取某个工作流的执行参数
        """
        pass

    @abstractmethod
    def plugin(self, plugin_name):
        """获取插件对象
        """
        pass


class AbstractWorkflow(object):

    """工作流抽象
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def add_listener(self, listener):
        """添加监听器
        """
        pass

    @abstractmethod
    def execute(self, start_point=None, args=None):
        """开始执行工作流
           :param start_point 指定运行的起始点
           :param args 工作流执行参数
           :return 返回一个End对象，描述执行结果
        """
        pass


class AbstractListener(object):

    """工作流监听器，此处定义了所有可监听的事件
       可以有选择感兴趣的事件进行监听
    """

    EVENTS = (
        "on_start",
        "on_unit_start",
        "on_unit_finish",
        "on_interrupt",
        "on_interrupt_now",
        "on_error",
        "on_finish"
    )

    @classmethod
    def wrap_function(cls, event_funcs):
        """将一个函数对象包装成一个监听器对象
           :param event_funcs 由事件名和函数对象所构成的元组
                  例如("on_start", start_func, "on_job_start", job_start_func)
        """
        return _WrappedFunctionListener(event_funcs)

    def __init__(self):
        pass

    def on_start(self, context):
        """在工作流调用了start方法之后立即进行
        """
        pass

    def on_unit_start(self, context):
        """在开始执行一个新的job时执行此方法
        """
        pass

    def on_unit_finish(self, context):
        """当一个Job被顺利完成时，执行此方法
           当因错误而结束时，不会执行此方法
        """
        pass

    def on_error(self, context,
                 exc_type, exc_value, traceback):
        """当工作流出现异常时，执行此方法
        """
        pass

    def on_finish(self, context):
        """当工作流顺利结束时，执行此方法
           当因错误而结束时，不会执行此方法
        """
        pass

    def __repr__(self):
        return "<Workflow event listener:{class_name} @ {id}>".format(
            class_name=self.__class__.__name__,
            id=id(self)
        )

    def __str__(self):
        return self.__repr__()


class _WrappedFunctionListener(AbstractListener):

    """用以包装函数的监听器类
    """

    def __init__(self, event_funcs):
        super(_WrappedFunctionListener, self).__init__()

        if len(event_funcs) % 2 != 0:
            raise InvalidArgumentException((
                u"event_funcs元数不正确，"
                u"必须为(事件名1, 函数1, 事件名2, 函数2, ...)的形式"))
        for event_name, func in itertools.izip(
                event_funcs[::2], event_funcs[1::2]):
            if not event_name.startswith("on_"):
                event_name = "on_" + event_name

            if event_name not in AbstractListener.EVENTS:
                raise InvalidArgumentException(u"未知的事件：{}".format(event_name))

            def event_handler_factory(func):
                def event_handler(ctx, f=func):
                    return f(ctx)

                return event_handler

            setattr(self, event_name, event_handler_factory(func))


class Env(object):

    """环境对象，用于描述工作流当前所运行的环境，比如测试环境、正式环境
       不同环境可携带不同的参数
    """

    @classmethod
    def test_env(cls):
        return cls("test", {}, Config(), u"测试环境")

    @args2fields(False)
    def __init__(self, name, args={}, config={}, description=""):
        """
        :param name 环境名称，比如test
        :param args 该环境的参数
        :param config 该环境当前的配置
        :param description 环境描述
        """
        pass



# coding: utf-8

import random
import webbrowser
from girlfriend.data.table import (
    Title,
    ListTable,
    TableWrapper
)
from girlfriend.plugin.table import (
    TableMeta,
    TableColumn2TitlePlugin,
    ZhMonthGenerator,
    PrintTablePlugin,
    ConcatTablePlugin,
    SplitTablePlugin,
    JoinTablePlugin,
    HTMLTable,
)
from girlfriend.testing import GirlFriendTestCase


class TableMetaTestCase(GirlFriendTestCase):

    def test_adapt_context_variable(self):
        """测试直接包装上下文变量
        """
        ctx = {
            "students": (
                (1, "Sam", 1),
                (2, "Jack", 2),
                (3, "James", 3)
            )
        }

        table_meta = TableMeta(
            "students",
            "students_table",
            u"学生表",
            ("id", u"编号", "name", u"姓名", "grade", u"年级")
        )

        table_meta(ctx)
        table = ctx["students_table"]
        self.assertEquals(len(table), 3)
        self.assertEquals(
            (table[1].id, table[1].name, table[1].grade),
            (2, "Jack", 2)
        )

    def test_adapt_list(self):
        """测试包装列表
        """

        ctx = {
            "total_students": 100,
            "new_students_by_year": 20,
            "jack_score": ["Jack", 120, 100, 101],
            "mike_score": ["Mike", 110, 109, 90]
        }

        table_meta = TableMeta(
            [("$total_students", "$new_students_by_year", 15, 5)],
            "summary_table",
            u"汇总表",
            (
                "total_students", u"学生总数",
                "new_students", u"新增学生数",
                "new_male", u"新增男生数",
                "new_female", u"新增女生数"
            )
        )

        table_meta(ctx)
        t = ctx["summary_table"]
        self.assertEquals(len(t), 1)
        self.assertEquals(
            (t[0].total_students, t[0].new_students,
             t[0].new_male, t[0].new_female),
            (100, 20, 15, 5)
        )

        table_meta = TableMeta(
            ["$jack_score", "$mike_score"],
            "score_table",
            u"成绩表",
            (
                "name", u"姓名",
                "chinese", u"语文",
                "maths", u"数学",
                "english", u"英语"
            )
        )

        table_meta(ctx)
        t = ctx["score_table"]
        self.assertEquals(len(t), 2)
        self.assertEquals(
            (t[0].name, t[0].chinese, t[0].maths, t[0].english),
            ("Jack", 120, 100, 101)
        )


class TableColumn2TitlePluginTestCase(GirlFriendTestCase):

    def setUp(self):
        self.ctx = {}
        titles = (
            Title("id", u"编号"),
            Title("name", u"姓名"),
            Title("grade", u"年级"),
            Title("month", u"月"),
            Title("score", u"得分")
        )

        students = (
            (1, "Sam", 1),
            (2, "James", 1),
            (3, "Joe", 2),
            (4, "Jack", 2)
        )

        data = []
        for i in xrange(0, len(students)):
            student = students[i]
            for m in xrange(1, 13):
                score = random.randint(0, 100)
                row = []
                row.extend(student)
                row.extend([m, score])
                data.append(row)
        random.shuffle(data)
        self.ctx["students_table"] = ListTable("students", titles, data)

    def test_transform(self):
        transformer = TableColumn2TitlePlugin()
        transformer.execute(
            self.ctx, "students_table", "changed_table",
            "month", "score", ZhMonthGenerator,
            sum_title=Title("sum", u"总分"),
            avg_title=Title("avg", u"平均分"))
        print "\nBefore change:"
        students_table = self.ctx["students_table"]
        print "\n"
        print students_table
        changed_table = self.ctx["changed_table"]
        print "\nAfter change:"
        print "\n"
        print changed_table


class PrintTablePluginTestCase(GirlFriendTestCase):

    def test_print(self):
        table = ListTable(
            "test",
            (
                Title("id", u"编号"),
                Title("name", u"姓名"),
                Title("grade", u"年级")
            ),
            (
                (1, "Sam", "A"),
                (2, "Tom", "A"),
                (3, "Jack", "B"),
                (4, "James", "C")
            )
        )

        pp = PrintTablePlugin()
        pp.execute(None, table, table, table)


class ConcatTablePluginTestCase(GirlFriendTestCase):

    def setUp(self):
        self.table_a = TableWrapper(
            "table_a",
            ("id", None, "name", None, "grade", None, "gender", None))((
                (1, "Sam", "A", "M"),
                (2, "Jack", "B", "M"),
                (3, "Lucy", "B", "F")
            ))

        self.table_b = TableWrapper(
            "table_b",
            ("id", None, "name", None, "grade", None, "gender", None))((
                (4, "James", "A", "M"),
                (5, "Peter", "A", "M")
            ))

        self.table_c = TableWrapper(
            "table_c", ("name", None, "grade", None, "gender", None))((
                ("Betty", "C", "F"),
                ("May", "C", "F")
            ))

    def test_concat(self):
        ctx = {}
        concat_table = ConcatTablePlugin()

        # titles is a number index
        table = concat_table.execute(
            ctx, (self.table_a, self.table_b),
            "concat_table"
        )

        print table

        self.assertEquals(len(table), 5)
        self.assertEquals(tuple(table[0]), (1, "Sam", "A", "M"))

        # 合并列数不相同的表
        table = concat_table.execute(
            ctx, (
                (self.table_a, "name", "grade", "gender"),
                (self.table_b, "name", "grade", "gender"),
                self.table_c
            ),
            "concat_table"
        )

        print table

        self.assertEquals(len(table), 7)
        self.assertEquals(tuple(table[0]), ("Sam", "A", "M"))


class JoinTablePluginTestCase(GirlFriendTestCase):

    def setUp(self):
        self.students_table = ListTable(
            "students",
            (Title("id", u"编号"), Title("name", u"姓名"),
             Title("grade", u"年级"), Title("city", u"城市"),
             Title("class", u"班级")),
            (
                (1, "Sam", 1, 1, 2),
                (2, "Jack", 1, 2, 3),
                (3, "Lucy", 2, 4, 1),
                (4, "James", 2, 3, 2),
                (5, "Larry", 2, 10, 3),
                (6, "Betty", 2, 10, 5),
            )
        )

        self.city_table = ListTable(
            "city",
            (Title("id", u"编号"), Title("name", u"名称")),
            (
                (1, "Pyongyang"),
                (2, "Beijing"),
                (3, "Shanghai"),
                (4, "Guangzhou"),
                (5, "Jinan")
            )
        )

        self.class_scores = ListTable(
            "class_score",
            (Title("grade", u"年级"), Title(
                "class", u"班级"), Title("score", u"分数")),
            (
                (1, 1, 10),
                (1, 2, 20),
                (1, 3, 30),
                (2, 1, 40),
                (2, 2, 50),
                (2, 3, 60)
            )
        )

    def test_inner_join(self):
        join_table = JoinTablePlugin()
        ctx = {}
        result = join_table.execute(
            ctx,
            "inner",
            self.students_table,
            self.city_table,
            on="city = id",
            fields=("l.id", "l.name", "r.name"),
            name=u"学生所在地",
            titles=(Title("id", u"编号"), Title(
                "name", u"姓名"), Title("city", u"城市"))
        )

        self.assertEquals(len(result), 4)
        row_0 = result[0]
        self.assertEquals(
            (row_0.id, row_0.name, row_0.city),
            (1, "Sam", "Pyongyang"),
        )

        print "\n"
        print result

        # mutli on conditions
        result = join_table.execute(
            ctx,
            "inner",
            self.students_table,
            self.class_scores,
            on="grade=grade,class=class",
            fields=("l.id", "l.name", "r.grade", "r.class", "r.score"),
            name=u"班级得分",
            titles=(Title("id", u"编号"), Title(
                "name", u"姓名"), Title("grade", u"年级"), Title("clazz", u"班级"),
                Title("score", u"评分"))
        )

        self.assertEquals(len(result), 5)
        row_0 = result[0]
        self.assertEquals(
            (row_0.id, row_0.name, row_0.grade, row_0.clazz, row_0.score),
            (1, "Sam", 1, 2, 20)
        )

        print "\n"
        print result

    def test_left_join(self):
        join_table = JoinTablePlugin()
        ctx = {}

        result = join_table.execute(
            ctx,
            "left",
            self.students_table,
            self.city_table,
            on="city=id",
            fields=("l.id", "l.name", "r.name"),
            name=u"学生所在地",
            titles=(Title("id", u"编号"), Title(
                "name", u"姓名"), Title("city", u"城市"))
        )

        self.assertEquals(len(result), 6)
        last_row = result[-1]
        self.assertEquals(
            (last_row.id, last_row.name, last_row.city),
            (6, "Betty", None)
        )

        print "\n"
        print result

        result = join_table.execute(
            ctx,
            "left",
            self.students_table,
            self.class_scores,
            on="grade=grade,class=class",
            fields=("l.id", "l.name", "l.grade", "l.class", "r.score"),
            name=u"班级得分",
            titles=(Title("id", u"编号"), Title(
                "name", u"姓名"), Title("grade", u"年级"),
                Title("clazz", u"班级"), Title("score", u"分数"))
        )

        self.assertEquals(len(result), 6)
        last_row = result[-1]
        self.assertEquals(
            (last_row.id, last_row.name, last_row.grade,
             last_row.clazz, last_row.score),
            (6, "Betty", 2, 5, None)
        )

        print "\n"
        print result


class SplitTablePluginTestCase(GirlFriendTestCase):

    def setUp(self):
        self.students_table = ListTable(
            "students",
            (
                Title("id", u"编号"),
                Title("name", u"姓名"),
                Title("grade", u"年级")
            ),
            (
                (1, "Sam", 1),
                (2, "Jack", 1),
                (3, "James", 1),
                (4, "Betty", 2),
                (5, "Tom", 2),
                (6, "Lucy", 2),
                (7, "Jason", 3),
            )
        )

    def test_split(self):
        ctx = {}
        split_table = SplitTablePlugin()

        def split_condition(row):
            return (
                "students_grade_{}".format(row.grade),
                u"{}年级学生".format(row.grade)
            )
        result = split_table.execute(ctx, self.students_table, split_condition)

        self.assertEquals(len(result["students_grade_1"]), 3)
        print "\n", result["students_grade_1"]
        self.assertEquals(len(result["students_grade_2"]), 3)
        print result["students_grade_2"]
        self.assertEquals(len(result["students_grade_3"]), 1)
        print result["students_grade_3"]


class HtmlTablePluginTestCase(GirlFriendTestCase):

    def setUp(self):
        self.table = ListTable(
            "student",
            (
                Title("id", u"编号"),
                Title("name", u"姓名"),
                Title("grade", u"年级")
            ),
            (
                (1, u"小王", 1),
                (2, u"小李", 1),
                (3, u"小明", 2)
            )
        )

        def data_row_style(row_index, row):
            if row.grade == 1:
                return "style='color:red;'"

        def data_cell_style(row_index, column_index, field_name, value):
            if field_name == "id":
                return "style='color:orange;'"

        def title_row_style(table):
            return "style='background-color:gray;'"

        def title_cell_style(column_index, title):
            if title.name == "name":
                return "style='text-decoration:underline;'"

        self.properties = {
            "table": "border=1",
            "title-row": title_row_style,
            "title-cell": title_cell_style,
            "data-row": data_row_style,
            "data-cell": data_cell_style,
        }

    def test_html_format(self):
        html_content = HTMLTable(
            self.table, "table", property=self.properties)({})
        with open("/tmp/table.html", "w") as f:
            f.write(html_content.encode("gbk"))
        webbrowser.open("file:///tmp/table.html")



# coding: utf-8

import argparse
import os.path
from girlfriend.plugin import Plugin, PluginManager
from girlfriend.workflow.gfworkflow import Job
from girlfriend.workflow.protocol import Env

# cmd args parser
cmd_parser = argparse.ArgumentParser(u"加1减2")
cmd_parser.add_argument("--start-num", dest="start_num",
                        help=u"起始值", default="0")

# 插件
plugins = [
    "multiply",
]
plugin_manager = PluginManager()


def multiply(ctx, a, b):
    return a * b
plugin_manager.register(
    Plugin.wrap_function("multiply", "multiply number", multiply))


def add(ctx, a, b):
    return a + b
plugin_manager.register(Plugin.wrap_function("add", "add number", add))


def minus(ctx, a, b):
    result = a - b
    with open(os.path.expanduser("~/test_workflow.txt"), "a") as f:
        f.write("{}\n".format(result))
    return result
plugin_manager.register(Plugin.wrap_function("minus", "minus number", minus))

# 配置
config = {
    "test": {
        "test_a_item": "a",
        "test_b_item": "b"
    }
}


def _test_args(cmd_options):
    return {
        "add_one": {
            "a": 1000,
        },
        "minus_two": {
            "a": "$add_one.result"
        }
    }


def _product_args(cmd_options):
    return {
        "add_one": {
            "a": 2000,
        },
        "minus_two": {
            "a": "$add_one.result"
        }
    }

# environments
env = (
    Env("test", _test_args),
    Env("product", _product_args),
)


def args(cmd_options):
    # args generator
    return {
        "add_one": {
            "a": int(cmd_options.start_num),
        },
        "minus_two": {
            "a": "$add_one.result"
        }
    }


def test_multiply(context):
    multiply_plugin = context.plugin("multiply")
    print multiply_plugin.execute(
        context,
        context["add_one.result"],
        context["minus_two.result"]
    )

# workflow units
workflow = (
    Job("add_one", "add", args={"b": 1}),
    Job("minus_two", "minus", args={"b": 2}),
    Job("multiply", caller=test_multiply)
)



# coding: utf-8

import re
import types
import prettytable
from collections import defaultdict
from girlfriend.data.table import (
    AbstractTable,
    TableWrapper,
    Title,
    ListTable
)
from girlfriend.util.lang import (
    args2fields,
    SequenceCollectionType
)
from girlfriend.exception import (
    InvalidTypeException,
    InvalidArgumentException
)


class TableAdapterPlugin(object):

    """将上下文中的二维列表、对象列表、字典列表适配成Table对象
       使用方法:

       Job("table_adapter", args=(
            TableMeta("test", "test_table", u"测试表",
                     (Title("id"), Title("name")))
       ))
    """

    name = "table_adapter"

    def __init__(self):
        pass

    def execute(self, context, *table_metas):
        return [table_meta(context) for table_meta in table_metas]


class TableMeta(object):

    """装载Table元数据，并对ctx_variable进行转换
    """

    @args2fields()
    def __init__(self, from_variable, to_variable,
                 name, titles, table_type=None):
        """
        :param from_variable 要转换为table对象的context变量名
        :param to_variable   转换结果的对象变量名
        :param name          表格名称
        :param titles        表格标题列表
        :param table_type    表格类型，默认为自动推断
        """
        pass

    def __call__(self, context):
        if isinstance(self._from_variable, types.StringTypes):
            # 如果是字符串，那么将作为上下文变量名，直接从context中获取变量
            obj = context[self._from_variable]
            table = TableWrapper(
                self._name,
                self._titles,
                self._table_type)(obj)
            context[self._to_variable] = table
            return table
        elif isinstance(self._from_variable, SequenceCollectionType):
            # 如果是List或者Tuple类型，那么从头构造数据
            table_data = []
            for row in self._from_variable:
                if isinstance(row, SequenceCollectionType):
                    row_data = []
                    for element in row:
                        if isinstance(element, types.StringTypes) \
                                and element.startswith("$"):
                            element = context[element[1:]]
                        row_data.append(element)
                    table_data.append(row_data)
                elif (isinstance(row, types.StringTypes) and
                      row.startswith("$")):
                    row = context[row[1:]]
                    table_data.append(row)
                else:
                    table_data.append(row)
            table = TableWrapper(
                self._name,
                self._titles,
                self._table_type
            )(table_data)
            context[self._to_variable] = table
            return table


class TableColumn2TitlePlugin(object):

    """该插件可以将table对象的某一列值转换成行，例如，我们从数据库中得到每人每周得分的结果，
       通常是这种结构:

       id week name grade score
        1   1   Jack  1    95
        1   2   Jack  1    97
        1   3   Jack  1    94
        1   4   Jack  1    87
        2   1   Sam   2    86
        2   2   Sam   2    87
        2   3   Sam   2    96

        但很多时候为了方便横向比较，我们最终通常需要这样的格式：

        id name grade week_1 week_2 week_3 week_4
         1  Jack  1     95     97     94     87
         2  Sam   2     86     87     96     99

        该插件可以帮助快速完成此类工作
    """

    name = "column2title"

    def execute(self, context, from_table, to_table,
                title_column, value_column, title_generator=Title,
                new_title_sort=sorted, new_table_name=None, default=None,
                sum_title=None, avg_title=None):
        """
        :param context  上下文对象
        :param from_table  要转换的表，可以是具体的table对象，也可以是上下文变量名
        :param to_table  转换结果所保存的变量名
        :param title_column  要转换为标题的列，比如上例中的week
        :param value_column  为新列提供数据的列，比如上例中的score
        :param title_generator 新列生成器，接受一个列值作为参数，返回Title对象
                         例如: lambda week: Title("week_" + week, u"星期" + week)
        :param new_title_sort 新标题排序，接受新标题值的列表，返回排序后的结果
        :param sum_title 接受一个Title对象，如果不为None，那么将用此列保存合计
        :param avg_title 平均值
        """
        if isinstance(from_table, types.StringTypes):
            from_table = context[from_table]

        title_column_values = set()

        # 将title_column和value_column之外的列作为唯一列
        unique_fields = tuple(title.name for title in from_table.titles
                              if title.name != title_column and
                              title.name != value_column)

        tmp_result = defaultdict(dict)
        for row in from_table:
            unique_columns = row[unique_fields]
            title_column_value = row[title_column]
            new_column_value = row[value_column]
            tmp_result[unique_columns][title_column_value] = new_column_value
            title_column_values.add(title_column_value)

        # 生成标题列表
        title_column_values = new_title_sort(title_column_values)
        titles = [title for title in from_table.titles
                  if title.name != title_column and
                  title.name != value_column]
        titles.extend(title_generator(v) for v in title_column_values)

        if sum_title:
            titles.append(sum_title)
        if avg_title:
            titles.append(avg_title)

        table_name = (from_table.name
                      if new_table_name is None else new_table_name)

        data = []
        for unique_columns in tmp_result:
            row = []
            row.extend(unique_columns)
            series = tuple(tmp_result[unique_columns].get(col, default)
                           for col in tmp_result[unique_columns])
            row.extend(series)
            sum_value = None
            if sum_title:
                sum_value = sum(series)
                row.append(sum_value)
            if avg_title:
                if sum_value is None:
                    sum_value = sum(series)
                avg_value = sum_value / len(series)
                row.append(avg_value)
            data.append(row)

        context[to_table] = ListTable(table_name, titles, data)


ZH_MONTHS = (u"一", u"二", u"三", u"四", u"五", u"六",
             u"七", u"八", u"九", u"十", u"十一", u"十二")


def ZhMonthGenerator(month):
    return Title("month_" + str(month), ZH_MONTHS[month - 1] + u"月")


class PrintTablePlugin(object):

    """以MySQL表格的样式批量打印Table对象
    """

    name = "print_table"

    def execute(self, context, *tables):
        for table in tables:
            if isinstance(table, types.StringTypes):
                table = context[table]
            # 居中显示表格名称
            print table.name.center(100, "-")
            print "\n"
            ptable = prettytable.PrettyTable([t.title for t in table.titles])
            for row in table:
                ptable.add_row(row)
            print str(ptable)
            print "\n\n"


class HTMLTablePlugin(object):

    name = "html_table"

    def execute(self, context, *table_formatters):
        return [formatter(context) for formatter in table_formatters]


class HTMLTable(object):

    """HTML格式化单元
    """

    @args2fields()
    def __init__(self, table, variable=None, property=None):
        pass

    def __call__(self, context):

        # table element
        html_buffer = [u"<table {prop}>".format(
            prop=self._extract_properties("table", value=self._table)
        )]

        # thead element
        html_buffer.append(u"<thead>")

        # thead -> tr
        html_buffer.append(
            u"<tr {prop}>".format(
                prop=self._extract_properties("title-row", value=self._table)
            )
        )

        # thead -> tr -> td
        for column_index, title in enumerate(self._table.titles):
            html_buffer.append(
                u"<th {prop}>{t}</th>".format(
                    t=title.title,
                    prop=self._extract_properties(
                        "title-cell", column_index=column_index, value=title)
                )
            )

        # close thead
        html_buffer.append(u"</tr></thead>")

        # tbody
        html_buffer.append(u"<tbody>")

        # tbody tr
        for row_index, row in enumerate(self._table):
            html_buffer.append(
                u"<tr {prop}>".format(
                    prop=self._extract_properties(
                        "data-row", row_index=row_index, value=row)
                )
            )
            for column_index, title in enumerate(self._table.titles):
                col_value = row[title.name]
                html_buffer.append(
                    u"<td {prop}>{col}</td>".format(
                        prop=self._extract_properties(
                            "data-cell", row_index=row_index,
                            column_index=column_index,
                            field_name=title.name,
                            value=col_value),
                        col=col_value
                    )
                )
            html_buffer.append(u"</tr>")

        # finish tbody
        html_buffer.append(u"</tbody></table>")
        html_content = "".join(html_buffer)
        if self._variable:
            context[self._variable] = html_content
        return html_content

    def _extract_properties(self, element_type, row_index=None,
                            column_index=None, field_name=None, value=None):
        """提取元素属性
        """
        if not self._property:
            return ""
        properties = self._property.get(element_type)
        if not properties:
            return ""

        if isinstance(properties, types.FunctionType):
            if element_type == "table":
                properties = properties(value)
            elif element_type == "title-row":
                properties = properties(value)
            elif element_type == "title-cell":
                properties = properties(column_index, value)
            elif element_type == "data-row":
                properties = properties(row_index, value)
            elif element_type == "data-cell":
                properties = properties(
                    row_index, column_index, field_name, value)

        if not properties:
            return ""
        return properties


class ConcatTablePlugin(object):

    """将多个表格按纵轴拼接在一起，类似于关系数据库中的union操作
    """

    name = "concat_table"

    def execute(self, context, tables, name, titles=0, variable=None):
        """
        :param context 上下文对象
        :param name    表名
        :param titles  标题，可以是Title对象列表，也可以是数字，如果是数字那么将
                       使用对应的拼接表标题
        :param tables  要拼接的表格，接受三种格式:
                       1. 具体的Table对象
                       2. Table对象的上下文变量名
                       3. 元组，第一个元素为Table对象，后面为要拼接的属性名
        """

        result = []
        for idx, table in enumerate(tables):
            fields = []  # 要拼接的字段
            if isinstance(table, types.StringTypes):
                table = context[table]
                fields = [title.name for title in table.titles]
            elif isinstance(table, AbstractTable):
                fields = [title.name for title in table.titles]
            elif isinstance(table, SequenceCollectionType):
                table, fields = table[0], table[1:]
                if isinstance(table, types.StringTypes):
                    table = context[table]
            else:
                raise InvalidTypeException

            for row in table:
                result.append(row[fields])

            if isinstance(titles, int) and titles == idx:
                titles = [
                    title for title in table.titles if title.name in fields]

        if variable is None:
            return ListTable(name, titles, result)
        else:
            context[variable] = ListTable(name, titles, result)


class JoinTablePlugin(object):

    """join tables like relation database
    """

    name = "join_table"

    def execute(self, context, way, left, right, on, fields,
                name, titles=None, variable=None):
        """
        :param context 上下文对象
        :param way join方式，允许inner、left和right三种join方式
        :param on  join条件，left_column = right_column，多个条件用逗号隔开
        :param fields 结果字段
        :param name 结果表格名称
        :param titles 结果表格标题
        :param variable 用于存储的上下文变量
        """

        if way not in ("inner", "left", "right"):
            raise InvalidArgumentException(
                u"不合法的join方式'{}'，只支持inner、left、right三种join方式".format(way))

        conditions = self._parse_conditions(on)

        if isinstance(left, types.StringTypes):
            left = context[left]
        if isinstance(right, types.StringTypes):
            right = context[right]

        if way == "inner":
            result = self._inner_join(left, right, fields, conditions)
        elif way == "left":
            result = self._side_join("left", left, right, fields, conditions)
        elif way == "right":
            result = self._side_join("right", right, left, fields, conditions)

        if titles is None:
            titles = tuple(Title(f.split(".")[1]) for f in fields)

        table = ListTable(name, titles, result)
        if variable:
            context[variable] = table
        else:
            return table

    def _parse_conditions(self, on_conditions):
        """解析on条件表达式"""
        return [
            _JoinCondition.parse(statement)
            for statement in on_conditions.split(",")]

    def _inner_join(self, left_table, right_table, fields, conditions):
        left_fields = [c.left_field for c in conditions]
        right_fields = [c.right_field for c in conditions]

        right_table_hash = self._table_hash(
            right_table, right_fields)
        result = []

        for row in left_table:
            hash_code = hash(row[left_fields])
            right_row_index_list = right_table_hash[hash_code]

            if not right_row_index_list:
                continue

            for right_row_index in right_row_index_list:
                right_row = right_table[right_row_index]
                if row[left_fields] == right_row[right_fields]:
                    result.append(self._build_row(fields, row, right_row))
                    break
        return result

    def _side_join(self, way, main_table, sub_table, fields, conditions):

        left_fields = [c.left_field for c in conditions]
        right_fields = [c.right_field for c in conditions]

        if way == "left":
            main_fields, sub_fields = left_fields, right_fields
        elif way == "right":
            main_fields, sub_fields = right_fields, left_fields

        sub_table_hash = self._table_hash(sub_table, sub_fields)

        result = []

        for row in main_table:
            hash_code = hash(row[main_fields])
            sub_row_index_list = sub_table_hash[hash_code]

            if way == "left":
                left_row, right_row = row, None
            else:
                left_row, right_row = None, row

            if not sub_row_index_list:
                result.append(self._build_row(fields, left_row, right_row))
                continue

            missed = True
            for sub_row_index in sub_row_index_list:
                sub_row = sub_table[sub_row_index]
                if row[main_fields] == sub_row[sub_fields]:
                    if way == "left":
                        left_row, right_row = row, sub_row
                    else:
                        left_row, right_row = sub_row, row
                    result.append(self._build_row(fields, left_row, right_row))
                    missed = False
                    break

            if missed:
                result.append(self._build_row(fields, left_row, right_row))

        return result

    def _table_hash(self, table, fields):
        hash_row_dict = defaultdict(list)
        for idx, row in enumerate(table):
            hash_row_dict[hash(row[fields])].append(idx)
        return hash_row_dict

    def _build_row(self, fields, left_row, right_row):
        if not fields:
            return tuple(left_row) + tuple(right_row)
        row = []
        for field in fields:
            if field.startswith(("r.", "right.")):
                if right_row:
                    row.append(right_row[field.split(".")[1]])
                else:
                    row.append(None)
            elif field.startswith(("l.", "left.")):
                if left_row:
                    row.append(left_row[field.split(".")[1]])
                else:
                    row.append(None)
        return row


class _JoinCondition(object):

    """Join 条件"""

    PATTERN = re.compile(r"\s*([\w_]+)\s*=\s*([\w_]+)\s*")

    @classmethod
    def parse(cls, statement):
        match_result = _JoinCondition.PATTERN.search(statement)
        if not match_result:
            raise InvalidArgumentException(u"表达式'{}'格式不合法".format(statement))
        return cls(match_result.group(1), match_result.group(2))

    def __init__(self, left_field, right_field):
        self.left_field = left_field
        self.right_field = right_field


class SplitTablePlugin(object):

    """分割表插件，根据指定条件将一个Table对象分割成多个Table对象"""

    name = "split_table"

    def execute(self, context, table, split_condition, variable=None):
        """
        :param context 上下文对象
        :param table 要分割的table对象
        :param split_condition 接受一个函数对象，参数为一个行对象，结果返回两个值，
                               第一个为引用的key，第二个为表格名称
        :param variable 要保存到的上下文变量

        :return 返回一个字典对象，key为split_condition中返回的引用值，value为表格对象
        """
        result = {}

        for row in table:
            split_result = split_condition(row)
            if split_result is None:
                continue
            ref_key, sub_table_name = split_result
            if ref_key in result:
                sub_table = result[ref_key]
                sub_table.append(row.obj)
            else:
                sub_table = TableWrapper(
                    sub_table_name,
                    table.titles
                )([row.obj])
                result[ref_key] = sub_table

        if variable:
            context[variable] = result

        return result



# coding: utf-8

"""本模块的目的是为了增加艺术性和情趣
"""

# 铃木爱理
MM_LINGMUAILI = r"""
XXXXXXXXXXXXXNXXXXNXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXKOc,...........,,''';cd0KXXXXXXXXXXXXXXXXXXNXXNNNNNNNNNNNNNNNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXKOo;......................';lxKXXXXXXXXXXXXXXXXNXXNNXNNNNNNNNNNNNNNNNNNNN
NXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXKKd;.............................':kKXXXXXXXXXXXXXXXXXXNXXNNNNXNNNNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXKd,..................................;kKXXXXXXXXXXXXXXXXNNNNNNNNNNNNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXKK0c.....................................':kXXXXXXXXXXXXXXXXXNNNXNNNNNNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXKKKKk;........................................'kKXXXXXXXXXXXXXNXXXXXNNNNNNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXKKKKKx'.........................................',kKXXXXXXXXXXXXXXXNNNXNNNNNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXKKKKKK0,.......''''...............................'':0KXXXXXXXXXXXXXXXXXNNNNNNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXKKKKKO,.......,''','.....':c......................'',oKXXXXXXXXXXXXXXXXNNNNNNNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXKKKKKKKc.......','',,''''';okx'..................'.'''';xKKXXXXXXXXXXXXXXXNNNNNNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXKKKKKKK0,.......''.'','''':dOOk;'.''''''''.'....'.'''',',lKKXXXXXXXXXXXXXXNXNNNNNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXKKKKKKK0;..........'''...;dO00Ol,'''''''''''''........'',,kKXXXXXXXXXXXXXXNNNNNNNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXKKKKKKKK0c..........''...,dO0000d,'.....................'''cKXXXXXXXXXXXXXXXXNNNNNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXXXXKKKKKKKK0O'...............oO00KKKO,........................'c0KXXXXXXXXXXXXXXXXNXNNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXXXXKKKKKKKK0l...............ck00KKKKKc........................'cKKXXXXXXXXXXXXXXXXXNNNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXXXKKKKKKKK0Oc;';...........,dkkkkkkkxo.........................;0KKXXXXXXXXXXXXXXNXNNNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXXXKKKKKKKK0lc.c:...........lO0KKKKK0Oko'.'';;'',;',;;,,........'OKKKXXXXXXXXXXXXXXXXNNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXXKKKKKKKK0xc:.od,.........;xkdll,,;cxkOd,:doxdoodc::;::cc,.....'OKKKKXXXXXXXXXXXXXXXXNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXKKKKKKKK00o,..oolc:,.....;x0OxxkollkOkO0OkKKK0O0kd:::xx:lxc,...'OKKKXKKXXXXXXXXXXXXXXNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXKKKKKK000kc...dxdxo;....,x0KKKKKKKKKKKXXKKXXKKKKKK0000OO00Od'..,OKKKKKKXXXXXXXXXXXXXXNNNNNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXXKKKKKK000o,...:xkxd;...,d0KKKXXXXXXXXXXXKKKKKKKKXXXKKKKKKK0k;..;0KKKKKKXXXXXXXXXXXXXXXXNNXNNNNN
XXXXXXXXXXXXXXXXXXXXXXXXKKKKKKK00k;.....,dkxc..'cO00KKKXXXNNXXXXKKKKXXKKKXXXXXXXKKK0x,..c0KKKKKKKXXXXXXXXXXXXXXXNXNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXKKKKKKKK00l'..''...:dl..,lkO00KKXXXXXXXK0000KXXK00KXXXXXXKKK0:..'c0KKKKKKKXXXXXXXXXXXXXXXXXNNNNNN
XXXXXXXXXXXXXXXXXXXXXXXKKKKKKK00Oc'..'.........'lxOO000KXXXXXK0000KXXXK00KXXXXXXKK0l...'l00KKKKKKKXXXXXXXXXXXXXXXXNXNNNN
XXXXXXXXXXXXXXXXXXXXXXKKKKKKK0000..,............:dkO0000KXXXXXK0OkO000OO0KXXXXXXKKx....,d000KKKKKKKXXXXXXXXXXXXXXXXNNNNN
XXXXXXXXXXXXXXXXXXXXXKKKKKKKK00x.................,lxOKK00KKKKKKKXKKKKKK0KKKKKKKKKKO,...';o0000KKKKKKXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXKKKKKKKK00kc.................;oOKKO000OxxOO000KK00000OO0KKKKO;.....;cO000KKKKKKXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXKKKKKKKK00Oo:,......;.........,oKXKOO00KK0OkxxkkOkxdxk0KKKK0x'......;cx000KKKKKKKXXXXXXXXXXXXXXXXXXNX
XXXXXXXXXXXXXXXXXXKKKKKKKK000kc;'........'.......:0KK0kk00KKKKK000000000KKKK0k:....'...,lo000KKKKKKKKXXXXXXXXXXXXXXXXXNX
XXXXXXXXXXXXXXXXKKKKKKKKK000Oocc'..'......'.....'kKKKOodkO0KKKKKKKKKKKKKKK0Ol........,,;doO00KKKKKKKKKXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXKKKKKKKKKK0000xllc'''.,...........oKXK0klloxO0KKKXXXXKKKKKKOl..........,;;ook00KKKKKKKKXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXKKKKKKKKKK00000Oodc;'....'.........;0XKKkoooodxkO0KKKKKKKKK0o'.........,,''':ox0KKKKKKKKKKXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXKKKKKKKKK000000kodc'......,...K0xllOKK0OddddxxxkkkOOOOOOOOxc..........',''..':k00KKKKKKKKKXXXXXXXXXXXXXXXXNX
XXXXXXXXXXXKKKKKKKKKK00000Oxdl;........:00000KKKKKKOddxxkkOOOOOOOOOOOko...........';,'...':dO00KKKKKKKXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXKKKKKKKKK00000OOo:,'........,O0KKKKXXXKKOxxkkkO000000000OOk,............''.'..,:oddxO0KKKKKKXXXXXXXXXXXXXXXXX
XXXXXXXXXXKKKKKKKKKKK000Oko,''........oO00OkkkO0KXXKkkOOO00KKKKKKK00O:'............,,'...;llxOkxxk0KKKKXXXXXXXXXXXXXXXXX
XXXXXXXXXXKKKKKKXXXXXKK0Ox'.''.......l000000KKKK0O00OK0O000KKKKKKKK0l;,.............,,...'dxkOOO0kdkKKKXXXXXXXXXXXXXXXXX
XXXXXXXXXKKXXXXNNNXXXKKO0l...'.......d00KKKK0KKKKKOOk0000K0KKKKKKK0dcc,..............''...okKXKKK00xdKKKXXXXXXXXXXXXXXXX
XXXXXXXKKXXXXNNNNNNNXXK00k...'.......kKKKK0OkkkkO0K0dxO0000KKKKKKKOcdc................'...'oOKNXXXK0kdKKKKXXXXXXXXXXXXXX
XXXXXXKKXXXXNXNXNNNNXXKK00c..........kKKKKK0OOOkkkO0OO00000000KKK0lld,........',...........':kKXXXXKKkkXXXXXXXXXXXXXXXXX
XXXXXKKKXXNNNNNNNNNNXXXKK0O;.........OKKKXX00K0O000KKXKK0OO00KKKK0:kl.........,:......';lll:',cx0KKKKKd0NNNNXXXXXXXXXXXX
XXXXKKKXNNNNXNNNNNXXNXXKKK0x'........kKKXXXK0KK00KKXXXK0OO0KKKKKK0co.........':'.....,cxKKK0lcoxodOKKKOxNNNNNNXXXXXXXXXX
XXXKKKKXXNXXNNNNNNNNNNXKKXK0:.......'O0KXXXXKKKKKKKKKK0OO0KKKKKKKO;'.........;....'..,l0KXXKxOK0OkdOKK0xNNNWNXWXXXXXXXXX
XXKKKKXXNNXXXXNNNNNNNNXK0OOkc......,x0KKKXXXXXKKKKK00O000KKKKKKKK0,..........'...''.;:kKXXKKO0XXK0Od0XOONNNWNNWXXXXXXXXX
XXKKKKXXNNXXNXXNNNXNNKOOOkxl;.....l0KKXXKKXXKK000OkkOO000KKKKKKK0k'....'.........''lol0KXXKXOKXNNNXxkKkNWWNNNWNXXXXXXXXX
XXKKKXXXNNXXNXXNNXXXKKKKkxdo,....dKKXXXXKKKKK000OkkOOO000KKKKK00Ok'..'''......,,.'lkkxKXXXXX0XNNNNN0x0XWWNNNNNNXXXXXXXXX
XKKKKXXXNNXXNXXNNXXNNKKOOkl,,'..:0KXXXXXXXXKK0OkkOOO000KKKKKKKK00k;...':;'....;;;,oOKOXXNNX0XNNNNNXkONWWWNNNNNNNXXXXXXXX
XKKKKXNXNNXNNXXNNNNNX0kkd:;:c...kKXXXXXXXXXKK0kOOO000KKKKKKKKKKKK0o....,;:,....;;:okKOXNNNKKNNNXXXXkKXWWNXNNNNNNNXXXXXXX
"""



# coding: utf-8

"""编写命令行脚本时使用的辅助工具
"""

import traceback
from termcolor import colored


def show_traceback_and_exit(msg=u"", color="red"):
    """当异常发生时显示堆栈并退出
    """
    print colored(msg, color)
    traceback.print_exc()
    exit(0)


def show_msg_and_exit(msg, color="red"):
    """显示消息并退出程序
    """
    print colored(msg, color)
    exit(0)



# coding: utf-8

"""GirlFriend需要很多第三方的代码支持，比如自定义的插件、工作流等等，
这些第三方的代码和项目往往都是固定的格式，因此完全可以通过填充模板的方式来减少工作量。
gf_gen就是这种根据据设定好的模板生成代码和项目结构的工具。
"""

import sys
import imp
import os
import os.path
import argparse
from girlfriend.util import script, cmdargs
from girlfriend.exception import GirlFriendBizException


def main():
    args, tpl_arg_list = parse_arguments()
    if not args.template:
        script.show_msg_and_exit(u"必须使用-t参数指定一个模板名称", "yellow")
    template_module = load_template(args.template)
    if template_module is None:
        script.show_msg_and_exit(u"找不到模板 '{}'".format(args.template))
    if not hasattr(template_module, "gen"):
        script.show_msg_and_exit(u"模板模块中必须包含一个gen函数")
    tpl_args = {}  # 执行模板生成所需要的参数
    tpl_parser = getattr(template_module, "parser", None)

    # 显示参数说明
    if args.show_args:
        if tpl_parser is not None:
            print u"模板 '{}' 的参数说明：".format(args.template)
            cmdargs.print_help(tpl_parser)
            exit(0)
        else:
            print u"模板 '{}' 不需要参数。".format(args.template)
            exit(0)

    if tpl_parser:
        tpl_args, _ = tpl_parser.parse_known_args(tpl_arg_list)
        tpl_args = tpl_args.__dict__

    try:
        template_module.gen(path=args.path, **tpl_args)
    except GirlFriendBizException as biz_e:
        script.show_msg_and_exit(unicode(biz_e))
    except Exception:
        script.show_traceback_and_exit()


def parse_arguments():
    parser = argparse.ArgumentParser(
        description=__doc__.decode("utf-8"))
    parser.add_argument("-t", dest="template", help=u"使用的模板名称")
    parser.add_argument("-p", dest="path", default=os.getcwd(),
                        help=u"生成项目的路径，默认为当前路径")
    parser.add_argument("--show-args", dest="show_args",
                        action="store_true", help=u"是否显示模板参数的解释")
    ns, tpl_args = parser.parse_known_args(sys.argv[1:])
    return ns, tpl_args


def load_template(template_name):
    file_name = template_name

    # 将点号转换为文件路径分隔符
    if "." in file_name:
        file_name = os.path.join(*file_name.split("."))

    if not file_name.endswith(".py"):
        file_name += ".py"

    # 组成完整的文件路径
    template_path = os.path.join(
        os.environ["HOME"], ".gf/project_templates", file_name)

    if not os.path.exists(template_path):
        return None
    return imp.load_source(template_name, template_path)

if __name__ == "__main__":
    main()



# coding: utf-8

"""CSV读取与写入插件
"""
from __future__ import absolute_import

import types
import csv
import StringIO
import requests
from girlfriend.util.lang import args2fields
from girlfriend.util.resource import HTTP_SCHEMA
from girlfriend.plugin.data import (
    AbstractDataReader,
    AbstractDataWriter
)


class CSVReaderPlugin(object):

    name = "read_csv"

    def execute(self, context, *csv_readers):
        return [reader(context) for reader in csv_readers]


class CSVR(AbstractDataReader):

    """
    CSV读单元，以单个CSV文件为单位
    """

    @args2fields()
    def __init__(self, path=None, content=None,
                 record_handler=None, record_filter=None, result_wrapper=None,
                 dialect="excel", variable=None):
        """
        :param path csv文件路径，可以是文件路径，也可以是url地址
        :param content csv数据内容，指定了content就不需要指定了path了，反之亦然
        :param record_handler 记录处理器，接受解析后的行对象
        :param record_filter 对解析后的行对象进行过滤
        :param result_wrapper 对最终结果的包装器
        :param dialect csv方言，可以是字符串也可以是csv.Dialect对象
        :param variable context中的引用变量名
        """
        pass

    def __call__(self, context):
        # get Dialect object
        if isinstance(self._dialect, types.StringTypes):
            dialect = csv.get_dialect(self._dialect)

        # get content
        content = self._content
        if self._content is None:
            if self._path.startswith(HTTP_SCHEMA):
                content = requests.get(self._path).text
            else:
                with open(self._path, "r") as f:
                    content = f.read()

        if isinstance(content, types.StringTypes):
            content = StringIO.StringIO(content)

        result = []
        csv_reader = csv.reader(content, dialect=dialect)
        for row in csv_reader:
            self._handle_record(row, result.append)

        return self._handle_result(context, result)


class CSVWriterPlugin(object):

    name = "write_csv"

    def execute(self, context, *csv_writers):
        return [writer(context) for writer in csv_writers]


class CSVW(AbstractDataWriter):

    """
    CSV写单元，以单个CSV文件为单位
    """

    @args2fields()
    def __init__(self, path, object, record_handler=None,
                 record_filter=None, dialect="excel"):
        """
        :param path 写入文件路径，可以为None，只将CSV内容保存到变量中
        :param object 将要转换为csv格式的可迭代对象
        :param record_handler 行对象转换，将记录转换为适合csv输出的格式
        :param record_filter 行过滤器，对行记录进行过滤
        :param dialect 方言，可接受字符串格式的方言名称或者Dialect对象
        """
        pass

    def __call__(self, context):

        # get dialect object
        if isinstance(self._dialect, types.StringTypes):
            dialect = csv.get_dialect(self._dialect)

        if self._path.startswith("memory:"):
            buffer_ = StringIO.StringIO()
            self._write_object(buffer_, dialect)
            buffer_.seek(0)
            context[self._path[len("memory:"):]] = buffer_
        else:
            with open(self._path, "w") as f:
                self._write_object(f, dialect)
        return self._path

    def _write_object(self, f, dialect):
        csv_writer = csv.writer(f, dialect=dialect)
        for row_obj in self._object:
            row_obj = self._handle_record(row_obj)
            if row_obj is None:
                continue
            csv_writer.writerow(row_obj)



# coding: utf-8

"""Plugin模块的测试用例
"""

import imp
import fixtures
from girlfriend.util.validating import Rule
from girlfriend.util.config import Config
from girlfriend.exception import InvalidArgumentException
from girlfriend.testing import GirlFriendTestCase
from girlfriend.plugin import (
    Plugin,
    PluginManager,
    PluginManagerChain,
    PluginAlreadyDeadException,
    PluginAlreadyPreparedException,
    PluginUnPreparedException,
    InvalidPluginException,
    PluginAlreadyRegisteredException,
    PluginNotFoundException,
    DefaultArgsValidator,
    DefaultConfigValidator,
)


class PluginTestCase(GirlFriendTestCase):

    """plugin.Plugin类的测试用例
    """

    def test_lifecycle(self):
        """测试Plugin的构造和生命周期
        """
        plugin = Plugin(
            name="testing",
            description="testing plugin",
            execute=lambda ctx, arg: "execute {}".format(arg),
            sys_prepare=lambda config: "sys_prepare",
            sys_cleanup=lambda config: "sys_cleanup",
            args_validator=None,
            config_validator=None
        )

        self.assertEquals(plugin.sys_prepare(None), "sys_prepare")
        self.assertEquals(plugin.execute(None, "cat"), "execute cat")
        self.assertEquals(plugin.sys_cleanup(None), "sys_cleanup")

        # 已经cleanup之后再去执行会抛出PluginAlreadyDeadException
        self.failUnlessException(
            PluginAlreadyDeadException,
            Plugin.execute,
            plugin, None, "cat")

        # 重复执行sys_prepare操作会抛出PluginAlreadyPreparedException
        self.failUnlessException(
            PluginAlreadyPreparedException,
            Plugin.sys_prepare,
            plugin, None)

        plugin = Plugin(
            name="testing",
            description="testing plugin",
            execute=lambda ctx, arg: "execute {}".format(arg),
            sys_prepare=lambda config: "sys_prepare",
            sys_cleanup=lambda config: "sys_cleanup",
            args_validator=None,
            config_validator=None
        )

        # 尚未执行初始化时执行会抛出PluginUnPreparedException
        self.failUnlessException(
            PluginUnPreparedException,
            Plugin.execute,
            plugin, None, "cat")

        # 尚未执行初始化时清理会抛出PluginUnPreparedException
        self.failUnlessException(
            PluginUnPreparedException,
            Plugin.sys_cleanup,
            plugin, None)

    def test_wrap_function(self):
        """测试wrap_function类方法
        """

        # 只包装一个execute函数的时候
        plugin = Plugin.wrap_function(
            "testing",
            "testing plugin",
            lambda ctx, arg: "execute {}".format(arg))
        self.failUnlessException(
            PluginUnPreparedException,
            Plugin.execute,
            plugin, None, "cat")
        plugin.sys_prepare(None)
        self.assertEquals(plugin.execute(None, "cat"), "execute cat")
        plugin.sys_cleanup(None)
        self.failUnlessException(
            PluginAlreadyDeadException,
            Plugin.execute,
            plugin, None)

        # 包装错误的函数对象
        self.failUnlessException(
            InvalidPluginException,
            Plugin.wrap_function,
            "testing",
            "testing plugin",
            lambda: "hello"
        )

        self.failUnlessException(
            InvalidPluginException,
            Plugin.wrap_function,
            "testing",
            "testing plugin",
            lambda ctx: "execute",
            lambda config, args: "sys_prepare"
        )

        plugin = Plugin.wrap_function(
            "testing",
            "testing plugin",
            lambda ctx, arg: "execute {}".format(arg),
            lambda config: "sys_prepare a = {}".format(config["plugin"]["a"]),
            lambda config: "sys_cleanup",
            (Rule("arg", required=True, type=str, min=3, max=10),),
            (Rule("plugin.a", required=True, type=int, min=1, max=3),)
        )

        cfg = Config({"plugin": {"a": -1}})
        self.failUnlessException(
            InvalidArgumentException, plugin.sys_prepare, cfg)
        cfg = Config({"plugin": {"a": 1}})
        self.assertEquals(plugin.sys_prepare(cfg), "sys_prepare a = 1")
        self.assertEquals(plugin.execute(None, "cat"), "execute cat")
        self.failUnlessException(
            InvalidArgumentException, plugin.execute, None, "c")
        self.assertEquals(plugin.sys_cleanup(None), "sys_cleanup")

    def test_wrap_class(self):
        """测试wrap_class类方法
        """

        class TestPlugin(object):

            """testing plugin"""

            def __init__(self, a, b):
                self.a = a
                self.b = b

            def execute(self, context, arg):
                return "execute {0} {1}".format(arg, self.a + self.b)

            def sys_prepare(self, config):
                return "sys_prepare"

            def sys_cleanup(self, config):
                return "sys_cleanup"

        plugin = Plugin.wrap_class(TestPlugin, a=1, b=2)

        self.assertEquals(plugin.name, "TestPlugin")
        self.assertEquals(plugin.description, "testing plugin")
        self.assertEquals(plugin.sys_prepare(None), "sys_prepare")
        self.assertEquals(plugin.execute(None, "cat"), "execute cat 3")
        self.assertEquals(plugin.sys_cleanup(None), "sys_cleanup")

        # 复制一个TestPlugin并替换掉execute方法,检测参数错误
        test_plugin = type("TestPluginA", tuple(), dict(TestPlugin.__dict__))

        def execute(self):
            return "execute"

        setattr(test_plugin, "execute", execute)
        self.failUnlessException(
            InvalidPluginException,
            Plugin.wrap_class,
            test_plugin, 1, 2
        )

        test_plugin = type("TestPluginB", tuple(), dict(TestPlugin.__dict__))

        def sys_prepare(self, config, x):
            return "sys_prepare"

        setattr(test_plugin, "sys_prepare", sys_prepare)
        self.failUnlessException(
            InvalidPluginException,
            Plugin.wrap_class,
            test_plugin, 1, 2
        )

        # 测试名称属性
        test_plugin = type("TestPluginC", tuple(), dict(TestPlugin.__dict__))
        test_plugin.name = "hehe"
        plugin = Plugin.wrap_class(test_plugin, 1, 2)
        self.assertEquals(plugin.name, "hehe")

    def test_wrap_module(self):
        """测试wrap_module类方法
        """
        plugin_module = imp.new_module("test_plugin")
        plugin_module.execute = lambda ctx, arg: "execute {}".format(arg)

        plugin = Plugin.wrap_module(plugin_module)
        self.assertEquals(plugin.name, "test_plugin")

        plugin.sys_prepare(None)
        self.assertEquals(plugin.execute(None, "cat"), "execute cat")

        plugin_module.execute = lambda: "execute"
        self.failUnlessException(
            InvalidPluginException,
            Plugin.wrap_module,
            plugin_module
        )


class PluginManagerTestCase(GirlFriendTestCase):

    """PluginManager类的测试用例
    """

    def setUp(self):
        self.plugin_manager = PluginManager()

        self.plugin_a = Plugin.wrap_function(
            "test_plugin_a", "testing plugin a", lambda ctx: "execute a")
        self.plugin_b = Plugin.wrap_function(
            "test_plugin_b", "testing plugin b", lambda ctx: "execute b")

        self.plugin_manager.register(self.plugin_a)
        self.plugin_manager.register(self.plugin_b)

    def test_register(self):
        """测试注册类对象
        """
        self.plugin_manager.sys_prepare(None, "test_plugin_a")

        plugin_a = self.plugin_manager["test_plugin_a"]
        self.assertEquals(plugin_a.execute(None), "execute a")

        # 未进行初始化插件
        plugin_b = self.plugin_manager["test_plugin_b"]
        self.failUnlessException(
            PluginUnPreparedException,
            Plugin.execute,
            plugin_b, None)

        # 测试重复注册
        self.failUnlessException(
            PluginAlreadyRegisteredException,
            PluginManager.register,
            self.plugin_manager, self.plugin_a
        )

    def test_remove(self):
        """测试将插件移除
        """
        del self.plugin_manager["test_plugin_a"]
        self.failUnlessException(
            PluginNotFoundException,
            PluginManager.remove,
            self.plugin_manager, "test_plugin_a")
        self.failUnlessException(
            PluginNotFoundException,
            PluginManager.plugin,
            self.plugin_manager, "test_plugin_a")

    def test_replace(self):
        """测试插件替换
        """
        plugin_a = Plugin.wrap_function(
            "test_plugin_a",
            "testing plugin a",
            lambda ctx: "execute new a")
        self.plugin_manager.replace(plugin_a)
        self.plugin_manager.sys_prepare(None)
        self.assertEquals(
            self.plugin_manager["test_plugin_a"].execute(None),
            "execute new a")


class PluginManagerChainTestCase(GirlFriendTestCase):

    """PluginManagerChain测试用例
    """

    def test_getitem(self):
        plugin_a = Plugin.wrap_function(
            "test_plugin_a",
            "testing plugin a",
            lambda ctx: "execute a"
        )

        plugin_b = Plugin.wrap_function(
            "test_plugin_b",
            "testing plugin b",
            lambda ctx: "execute b"
        )

        plugin_manager_a = PluginManager()
        plugin_manager_a.register(plugin_a)
        plugin_manager_a.register(plugin_b)

        plugin_a1 = Plugin.wrap_function(
            "test_plugin_a",
            "testing plugin a 1",
            lambda ctx: "execute a 1"
        )

        plugin_manager_b = PluginManager()
        plugin_manager_b.register(plugin_a1)

        plugin_mgr_chain = PluginManagerChain(
            plugin_manager_b, plugin_manager_a)
        plugin_mgr_chain.sys_prepare(None)

        # 优先获取最优先的插件
        self.assertEquals(
            plugin_mgr_chain["test_plugin_a"].execute(None), "execute a 1")
        self.assertEquals(
            plugin_mgr_chain["test_plugin_b"].execute(None), "execute b")

        self.failUnlessException(
            PluginNotFoundException,
            PluginManagerChain.plugin,
            plugin_mgr_chain, "test_plugin_c")


# Plugin 相关Fixture

class PluginMgrFixture(fixtures.Fixture):

    """创建一个新的PluginManager并添加几个简单插件
    """

    def __init__(self, config, *plugins):
        if plugins:
            self._plugins = plugins
        else:
            self._plugins = [
                Plugin.wrap_function(
                    "plugin_a", "plugin a", lambda ctx: "plugin a"),
                Plugin.wrap_function(
                    "plugin_b", "plugin b", lambda ctx: "plugin b")
            ]
        self._config = config

    @property
    def plugin_manager(self):
        return self._plugin_manager

    def setUp(self):
        self._plugin_manager = PluginManager()
        for plugin in self._plugins:
            self._plugin_manager.register(plugin)
        self._plugin_manager.sys_prepare(self._config)

    def cleanUp(self):
        self._plugin_manager.sys_cleanup(self._config)


class DefaultArgsValidatorTestCase(GirlFriendTestCase):

    """默认参数验证器测试用例
    """

    def test_validate(self):
        rules = (
            Rule("id", type=int, required=True, min=10000, max=20000),
            Rule("name", type=str, required=False, min=2, max=6),
            Rule("email", type=str, required=True, logic=lambda value:
                 u"existed!" if value == "chz@gmail.com" else None)
        )
        validator = DefaultArgsValidator(rules)

        validator(15000, "SamChi", "sam@163.com")
        validator(15000, name=None, email="sam@163.com")

        self.failUnlessException(InvalidArgumentException,
                                 validator, 1, "SamChi", "aaaa")
        self.failUnlessException(InvalidArgumentException,
                                 validator, id=1, name=None, email="a")
        self.failUnlessException(InvalidArgumentException,
                                 validator, 10000, None, "chz@gmail.com")


class DefaultConfigValidatorTestCase(GirlFriendTestCase):

    """默认配置验证器测试用例
    """

    def test_validate(self):
        rules = (
            Rule("smtp.host", type=str, required=True),
            Rule("smtp.port", type=int, required=False, min=25, max=65535),
            Rule("smtp.user", type=str, required=True,
                 min=1, max=10, regex=r"^\w+$"),
            Rule("smtp.password", type=str, required=True, min=6, max=20),
            Rule("db.host", type=str, required=True, logic=lambda value:
                 None if value.startswith("192.168") else "Invalid!")
        )
        validator = DefaultConfigValidator(rules)

        cfg = Config({
            "smtp": {
                "host": "192.168.100.2",
                "port": 25,
                "user": "SamChi",
                "password": "123456"
            },
            "db": {
                "host": "192.168.1.1"
            }
        })
        validator(cfg)

        cfg["db"]["host"] = "localhost"
        self.failUnlessException(InvalidArgumentException, validator, cfg)



# coding: utf-8

import os
import xlrd
import os.path
from girlfriend.data.table import ListTable, Title
from girlfriend.testing import GirlFriendTestCase
from girlfriend.plugin.excel import (
    ExcelWriterPlugin,
    SheetW,
    CellStyle,
    ExcelReaderPlugin,
    SheetR,
)


class ExcelWriterPluginTestCase(GirlFriendTestCase):

    def setUp(self):
        self.students_list = (
            (1, u"迟宏泽", u"男", 1),
            (2, "Jack", u"男", 2),
            (3, u"韩梅梅", u"女", 3),
            (5, u"李雷", u"男", 3)
        )
        self.students_table = ListTable(
            "students",
            (
                Title("id", u"编号"),
                Title("name", u"姓名"),
                Title("gender", u"性别"),
                Title("grade", u"年级")
            ),
            self.students_list
        )

        def even_row(row_index, column_index):
            return row_index != 0 and row_index % 2 == 0

        self.even_row_style = CellStyle(even_row, {"font_color": "red"})
        self.top_three_style = CellStyle(
            ((1, 3), None), {"bg_color": "yellow"})

        self.tmp_path = []

    def test_write_table(self):
        ctx = {}
        excel_writer = ExcelWriterPlugin()

        # write to tmp
        path = excel_writer.execute(
            ctx, None,
            sheets=(SheetW(self.students_table),)
        )
        workbook = xlrd.open_workbook(path)
        self.tmp_path.append(path)
        self._check_table_workbook(workbook)

        # write to StringIO
        excel_writer.execute(
            ctx, "$students_xlsx",
            sheets=(SheetW(self.students_table),)
        )
        book_buffer = ctx["students_xlsx"]
        book_buffer.seek(0)
        workbook = xlrd.open_workbook(file_contents=book_buffer.buf)
        self._check_table_workbook(workbook)

        # write to target file

        def draw_chart(workbook, sheet):
            """绘制图表
            """
            chart = workbook.add_chart({"type": "column"})
            chart.add_series({
                "categories": "=students! $B$2: $B$5",
                "values": "=students! $D$2: $D$5",
            })
            sheet.insert_chart("A6", chart)

        path = excel_writer.execute(
            ctx, os.path.expanduser("~/Desktop/hehe2.xlsx"),
            sheets=(
                SheetW(
                    self.students_table, style=[
                        CellStyle(0, {"bold": True}),
                        self.even_row_style,
                        self.top_three_style
                    ],
                    sheet_handler=draw_chart
                ),
            ))
        workbook = xlrd.open_workbook(path)

        # write sequence
        excel_writer.execute(
            ctx, "$students_xlsx",
            sheets=(SheetW(self.students_list, sheet_name="std"),))
        book_buffer = ctx["students_xlsx"]
        book_buffer.seek(0)
        workbook = xlrd.open_workbook(file_contents=book_buffer.buf)
        student_table = workbook.sheet_by_name("std")
        self.assertEquals(student_table.nrows, len(self.students_list))

    def _check_table_workbook(self, workbook):
        self.assertEquals(len(workbook.sheets()), 1)
        student_table = workbook.sheet_by_name("students")
        self.assertEquals(
            student_table.nrows, len(self.students_table) + 1)
        self.assertEquals(
            student_table.ncols, len(self.students_table.titles))

    def tearDown(self):
        for path in self.tmp_path:
            os.remove(path)


class ExcelReaderPluginTestCase(GirlFriendTestCase):

    def setUp(self):
        self.students = (
            (1, "Sam", 1),
            (2, "Jack", 2),
            (3, "James", 2),
            (4, "Peter", 1)
        )
        self.students_table = ListTable(
            "students",
            (
                Title("id", u"编号"),
                Title("name", u"姓名"),
                Title("grade", u"年级")
            ),
            self.students
        )

        self.scores = (
            (1, 100, 98, 88),
            (2, 95, 99, 99),
            (3, 96, 97, 89),
            (4, 98, 97, 99)
        )
        self.score_table = ListTable(
            "students_score",
            (
                Title("id", u"编号"),
                Title("c", u"C语言"),
                Title("java", u"Java程序设计"),
                Title("ds", u"数据结构")
            ),
            self.scores
        )

        excel_writer = ExcelWriterPlugin()
        self.path = excel_writer.execute({}, None, sheets=(
            SheetW(self.students_table),
            SheetW(self.score_table)
        ))

    def test_read_table(self):
        excel_reader = ExcelReaderPlugin()
        ctx = {}
        excel_reader.execute(
            ctx, self.path,
            SheetR("students", variable="students", skip_first_row=True),
            SheetR("students_score", variable="scores", skip_first_row=True)
        )
        students, scores = ctx["students"], ctx["scores"]
        self.assertEquals(len(students), len(self.students))
        self.assertEquals(len(scores), len(self.scores))

    def tearDown(self):
        os.remove(self.path)



# coding: utf-8

"""配置解析相关
"""

import os
import os.path
import StringIO
import requests
from ConfigParser import SafeConfigParser
from girlfriend.exception import GirlFriendBizException

# 配置文件样例
CONFIG_EXAMPLE = """
[db_test]
connect_url=sqlite:///:memory:
"""


class Config(dict):

    """Config是一个Mapping实现，用于保存配置信息
    """

    @staticmethod
    def load_by_name(name="default"):
        """根据配置资源名称来加载配置数据
           :param name 资源名称，默认为default
        """
        default_config_path = os.path.join(
            os.environ["HOME"],
            ".gf",
            "gf.cfg"
        )
        if not os.path.exists(default_config_path):
            raise ConfigFileNotExistException(
                u"用户目录下的默认配置文件不存在，无法获取配置资源名称映射，"
                u"您可以尝试使用gf_config命令来初始化默认配置文件"
            )
        default_config = Config.load_from_file(default_config_path)
        if name == "default":
            return default_config
        config = default_config.get("config")
        if not config:
            raise UnknownConfigResourceException(
                u"找不到名称为{}的配置信息，请检查默认配置".format(name)
            )
        resource_path = config.get(name)
        if not resource_path:
            raise UnknownConfigResourceException(
                u"找不到名称为 '{}' 的配置信息，请检查默认配置".format(name)
            )
        if resource_path.startswith(("http://", "https://")):
            return Config.load_from_web(resource_path)
        else:
            return Config.load_from_file(resource_path)

    @staticmethod
    def load_from_file(file_path):
        """从文件中加载配置
           :param 文件路径
        """
        with open(file_path, "r") as f:
            return Config.wrap_fp(f)

    @staticmethod
    def load_from_web(url):
        """从web加载配置
           :param 会对该地址进行GET请求，请求获取的数据将作为配置文件
        """
        response = requests.get(url)
        return Config.wrap_fp(StringIO.StringIO(response.text))

    @classmethod
    def wrap_fp(cls, fp):
        """将一个文件描述符包装成Config对象
        """
        parser = SafeConfigParser()
        parser.readfp(fp)
        config = cls()
        for section_name in parser.sections():
            config[section_name] = cls(
                {name: value for name, value in parser.items(section_name)}
            )
        return config

    def __getattr__(self, key):
        if key not in self:
            raise AttributeError(u"属性 '{}' 不存在".format(key))
        return self[key]

    def __setattr__(self, key, value):
        self[key] = value

    def prefix(self, *pre):
        """根据Section前缀获取子配置项
        """
        return {section: self[section]
                for section in self if section.startswith(pre)}

    def get(self, section, item_name=None):
        if item_name is None:
            return dict.get(self, section)
        section_items = dict.get(self, section, {})
        if not section_items:
            return None
        return section_items.get(item_name)


class ConfigFileNotExistException(GirlFriendBizException):

    """当目标配置文件不存在时抛出此异常
    """
    pass


class UnknownConfigResourceException(GirlFriendBizException):

    """根据名称找不到对应的配置资源时抛出此异常
    """
    pass



# coding: utf-8

import os
import httpretty
from girlfriend.testing import GirlFriendTestCase
from girlfriend.plugin.csv import CSVR, CSVW


STUDENTS = (
    ("1", "Sam", "1"),
    ("2", "Jack", "2"),
    ("3", "James", "2"),
    ("4", "Lucy", "2"),
    ("5", "Peter", "1")
)

CSV_CONTENT = "\n".join(",".join(row) for row in STUDENTS)


class CSVReaderTestCase(GirlFriendTestCase):

    def setUp(self):
        with open("test.csv", "w") as f:
            f.write(CSV_CONTENT)

        httpretty.enable()

        httpretty.register_uri(
            httpretty.GET,
            "http://test.gf/csv",
            body=CSV_CONTENT,
            content_type="text/plain"
        )

    def test_read_from_file(self):
        result = CSVR(
            "test.csv", record_handler=tuple, result_wrapper=tuple)({})
        self.assertEquals(result, STUDENTS)
        result = CSVR(
            content=CSV_CONTENT, record_handler=tuple,
            result_wrapper=tuple)({})
        self.assertEquals(result, STUDENTS)

    def test_read_from_web(self):
        result = CSVR(
            "http://test.gf/csv",
            record_handler=tuple, result_wrapper=tuple)({})
        self.assertEquals(result, STUDENTS)

    def tearDown(self):
        os.remove("test.csv")
        httpretty.disable()
        httpretty.reset()


class Student(object):

    def __init__(self, id_, name, grade):
        self.id = id_
        self.name = name
        self.grade = grade

student_objects = (
    Student(1, "Sam", 1),
    Student(2, "Jack", 1),
    Student(3, "Tom", 2),
    Student(4, "Lucy", 2),
    Student(5, "James", 2)
)


class CSVWriterPluginTestCase(GirlFriendTestCase):

    def setUp(self):
        pass

    def test_write_file(self):
        CSVW(
            "test.csv", student_objects,
            record_handler=lambda std: (std.id, std.name, std.grade)
        )({})
        with open("test.csv", "r") as f:
            print "\n", f.read()
        os.remove("test.csv")

    def test_write_context_var(self):
        ctx = {}
        CSVW(
            "memory:test_csv", student_objects,
            record_handler=lambda std: (std.id, std.name, std.grade)
        )(ctx)
        print "\n", ctx["test_csv"].buf



# coding: utf-8

"""很多时候我们经常需要定义目录文件的生成模板，比如生成一个配置文件的模板，
生成一个工程目录结构等等，该模块用于帮助轻松实现这类工作。
"""

import os
import os.path
from abc import (
    ABCMeta,
    abstractproperty,
    abstractmethod
)
from girlfriend.exception import GirlFriendBizException


class TemplateUnit(object):

    """模板单元抽象
    """

    __metaclass__ = ABCMeta

    TYPE_FILE = "file"

    TYPE_DIR = "dir"

    def __init__(self, name, access=None):
        """
          :param name 目录名或文件名
          :param access  文件权限，默认按照umask来，接受一个八进制数字，比如0666
        """
        self._name = name
        self._access = access

    @property
    def name(self):
        """单元名称
        """
        return self._name

    @property
    def access(self):
        return self._access

    @abstractproperty
    def unittype(self):
        """单元类型
        """
        pass

    @abstractmethod
    def makeme(self, base_dir=os.getcwd()):
        """创建该单元及其子元素
        """
        pass


class Dir(TemplateUnit):

    def __init__(self, name, access=None, elements=None):
        super(Dir, self).__init__(name, access)
        if elements:
            self.elements = elements
        else:
            self.elements = []

    @property
    def unittype(self):
        return TemplateUnit.TYPE_DIR

    def append(self, element):
        self.elements.append(element)

    def makeme(self, base_dir=os.getcwd()):
        # 先建自己
        myself_dir = os.path.join(base_dir, self.name)
        if os.path.exists(myself_dir):
            raise DirAlreadyExistException(
                u"目录 '{}' 已经存在了，不能重复创建".format(myself_dir))
        os.mkdir(myself_dir)
        if self.access:
            os.chmod(myself_dir, self.access)
        # 创建子元素
        for element in self.elements:
            element.makeme(myself_dir)


class File(TemplateUnit):

    def __init__(self, name, access=None, content=""):
        super(File, self).__init__(name, access)
        self._content = content

    @property
    def unittype(self):
        return TemplateUnit.TYPE_FILE

    def makeme(self, base_dir):
        myself_path = os.path.join(base_dir, self.name)
        if os.path.exists(myself_path):
            raise FileAlreadyExistException(
                u"文件 '{}' 已经存在了，不能重复创建".format(myself_path))
        with open(myself_path, "w") as f:
            f.write(self._content)
        if self.access:
            os.chmod(myself_path, self.access)


class DirAlreadyExistException(GirlFriendBizException):

    """当目录已经存在时，抛出此异常
    """
    pass


class FileAlreadyExistException(GirlFriendBizException):

    """当文件已经存在时，抛出此异常
    """
    pass



# coding: utf-8

"""日志工具模块
   集成了常用的log handler、log formatter和日志创建工具
"""

import sys
import logging
import logging.handlers


LOGGER_LEVELS = {
    "critical": logging.CRITICAL,
    "error": logging.ERROR,
    "warning": logging.WARNING,
    "info": logging.INFO,
    "debug": logging.DEBUG,
}


def get_logger_level_by_name(name):
    return LOGGER_LEVELS[name.lower()]


def time_rotating_handler(filename, when="D", interval=1, backupCount=0,
                          encoding="utf-8", delay=False):
    """基于时间回滚的日志策略"""
    return logging.handlers.TimedRotatingFileHandler(
        filename,
        when="D",
        interval=1,
        backupCount=0,
        encoding="utf-8",
        delay=delay
    )


def daily_rotaiting_handler(filename):
    """基于天的日志回滚策略"""
    return time_rotating_handler(filename)


def stdout_handler():
    """标准输出"""
    return logging.StreamHandler(sys.stdout)


LOG_FORMATTER = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"


def create_logger(logger_name, handlers, level=logging.INFO,
                  formatter=LOG_FORMATTER):
    """创建日志对象
    :param logger_name 日志名称
    :param filename 日志文件路径
    :param level 日志级别
    :param formatter 日志输出格式
    :param handlers 日志处理策略
    """
    logger = logging.getLogger(logger_name)
    logger.setLevel(level)
    for handler in handlers:
        handler.setFormatter(logging.Formatter(LOG_FORMATTER))
        logger.addHandler(handler)
    return logger



# coding: utf-8

import os
from StringIO import StringIO
from email.utils import COMMASPACE
from girlfriend.testing import GirlFriendTestCase
from girlfriend.plugin.mail import (
    SMTPManager,
    Attachment,
    SendMailPlugin,
    Mail,
)
from girlfriend.util.config import Config


class SMTPManagerTestCase(GirlFriendTestCase):

    def test_validate_config(self):
        smtp_manager = SMTPManager()
        config = Config({
            "smtp_test": {
                "host": "smtp.163.com",
                "account": "17600817832@163.com",
                "password": "gf123456"
            }
        })
        smtp_manager.validate_config(config)
        self.assertEquals(config["smtp_test"]["port"], 25)

        config["smtp_test"]["port"] = "25"
        smtp_manager = SMTPManager()
        smtp_manager.validate_config(config)
        self.assertEquals(config["smtp_test"]["port"], 25)


class AttachmentTestCase(GirlFriendTestCase):

    def test_build_mime_object(self):
        with open("test_attachment.txt", "w") as f:
            f.write("Hello! Hawaii!")

        # test file path
        attachment = Attachment("test_attachment.txt",
                                "text/plain", u"测试文本.txt".encode("utf-8"))
        mime_object = attachment.build_mime_object()

        # test file object
        attachment = Attachment(
            open("test_attachment.txt", "r"), "text/plain",
            u"测试文本.txt".encode("utf-8"))
        mime_object_2 = attachment.build_mime_object()
        self.assertEquals(mime_object.as_string(), mime_object_2.as_string())

        # test StringIO
        attachment = Attachment(
            StringIO("Hello! Hawaii!"), "text/plain",
            u"测试文本.txt".encode("utf-8"))
        mime_object_3 = attachment.build_mime_object()
        self.assertEquals(mime_object.as_string(), mime_object_3.as_string())

        os.remove("test_attachment.txt")


class SendMailPluginTestCase(GirlFriendTestCase):

    def setUp(self):
        self.config = Config({
            "smtp_test": {
                "host": "smtp.163.com",
                "port": 465,
                "account": "17600817832@163.com",
                "password": "gf123456",
                "ssl": "true"
            }
        })
        self.send_mail_plugin = SendMailPlugin()
        self.send_mail_plugin.config_validator(self.config)

    def test_execute(self):
        ctx = {}
        self.send_mail_plugin.execute(
            ctx,
            "test",
            receivers="17600817832@163.com",
            sender="17600817832@163.com",
            subject=u"新年快乐",
            content=u"新年快乐!"
        )

    def test_multi_receiver(self):
        ctx = {}
        self.send_mail_plugin.execute(
            ctx,
            "test",
            receivers=COMMASPACE.join([
                "17600817832@163.com",
                "chihz3800@163.com",
                "hongze.chi@gmail.com"
            ]),
            sender="17600817832@163.com",
            subject=lambda ctx, receiver: u"你好, " + receiver,
            content=lambda ctx, receiver: u"你好, " + receiver
        )

        self.send_mail_plugin.execute(
            ctx,
            "test",
            receivers=[
                "hongze.chi@gmail.com",
                "chihz3800@163.com",
            ],
            sender="17600817832@163.com",
            subject=lambda ctx, receiver: u"Hahaha, " + receiver,
            content=lambda ctx, receiver: u"你好, content " + receiver,
            attachments=[
                Attachment(StringIO("simple!"),
                           "text/plain", u"文本.txt".encode("gb2312")),
                Attachment(StringIO("naive!"),
                           "text/plain", u"文本2.txt".encode("gb2312"))
            ]
        )

    def test_customize(self):

        class MyMail(Mail):

            def __init__(self, context, receiver):
                super(MyMail, self).__init__(context, receiver)

            @property
            def sender(self):
                return "17600817832@163.com"

            @property
            def receiver_email(self):
                return self._receiver.email

            @property
            def subject(self):
                return u"新年快乐"

            @property
            def content(self):
                return u"新年快乐，发大财！{}".format(self._receiver.name)

            @property
            def attachments(self):
                return [
                    Attachment(StringIO("simple!"),
                               "text/plain", u"文本.txt".encode("gb2312")),
                    Attachment(StringIO("naive!"),
                               "text/plain", u"文本2.txt".encode("gb2312"))
                ]

        class User(object):

            def __init__(self, name, email):
                self.name = name
                self.email = email

        ctx = {}
        self.send_mail_plugin.execute(
            ctx,
            "test",
            receivers=[
                User(u"迟宏泽", "h.ongzechi@gmail.com"),
                User(u"小白", "hongze.chi@gmail.com")
            ],
            mail=MyMail
        )



# coding: utf-8

"""基于SQLAlchemy的ORM插件
"""

import re
import ujson
import types
import threading
import collections
from sqlalchemy import create_engine, text
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.query import Query as ORMQuery
from sqlalchemy.pool import Pool, QueuePool, NullPool
from sqlalchemy.ext.automap import automap_base
from girlfriend.exception import InvalidStatusException
from girlfriend.util.validating import Rule, be_json
from girlfriend.util.lang import (
    args2fields,
    SequenceCollectionType,
    parse_context_var
)


class EngineManager(object):

    """通过EngineManager来管理所有的Engine对象
       这样所有的SQLAlchemy插件就可以实现Engine共享
    """

    STATUS_UNINIT = 0  # 尚未初始化

    STATUS_OK = 1  # 状态OK

    STATUS_DISPOSED = 2  # 已经销毁

    # 数据源配置项的验证规则
    config_rules = (
        # 连接字符串，比如postgresql://scott:tiger@localhost/test
        Rule("connect_url", required=True,
             type=types.StringTypes, regex=r"^\w+://"),
        # 编码
        Rule("encoding", required=False,
             type=types.StringTypes, default="utf-8"),
        # 连接参数，采用json的形式:{'ssl': {'cert': xxx, 'key': xxx, 'ca': xxx}}
        Rule("connect_args", required=False,
             type=types.StringTypes, default="{}",
             logic=be_json("connect_args")),
        # 该数据源使用的连接池策略
        Rule("pool_policy", required=False,
             type=types.StringTypes, default=None)
    )

    # 连接池的配置项验证规则，section名称皆以dbpool_开头
    pool_config_rules = (
        # 连接池类型，比如sqlalchemy.pool.QueuePool，可以是字符串，也可以是具体的类对象
        Rule("poolclass", required=True, type=(Pool, types.StringTypes)),
        # 连接池大小
        Rule("pool_size", required=False, type=(int, types.StringTypes),
             regex=r"^\d+$", default=10),
        # 连接回收周期，单位为秒，MySQL尤其要注意设置此项，否则连接会自动过期断开
        Rule("pool_recycle", required=False, type=(int,),
             regex=r"^\d+$", default=3600),
        # 从连接池中获取连接的超时时间，SQLAlchemy自带连接池中，只能用于QueuePool
        Rule("pool_timeout", required=False, type=(
            int,), regex=r"^\d+$", default=30),
    )

    def __init__(self):
        self._engines = {}
        self._status = EngineManager.STATUS_UNINIT
        self._validated = False

    def validate_config(self, config):
        """统一对配置进行验证，避免各个插件单独验证
        """
        if self._validated:
            return
        for section in config.prefix("db_", "dbpool_"):
            if section.startswith("db_"):
                self._validate_config_items(
                    config, section, EngineManager.config_rules)
            elif section.startswith("dbpool_"):
                self._validate_config_items(
                    config, section, EngineManager.pool_config_rules)

        self._validated = True

    def _validate_config_items(self, config, section, rules):
        config_items = config[section]
        for rule in rules:
            item_value = config_items.get(rule.name)
            rule.validate(item_value)
            if item_value is None and not rule.required:
                config[section][rule.name] = rule.default

    def init_all(self, config):
        """统一初始化
        """
        if self._status != EngineManager.STATUS_UNINIT:
            return
        all_pool_config = self._load_db_pool_config(config)
        for section in config.prefix("db_"):
            engine_name = section.split("_", 1)[1]
            engine = self._create_engine(config, section, all_pool_config)
            self._engines[engine_name] = EngineContainer(engine)
        self._status = EngineManager.STATUS_OK

    def _load_db_pool_config(self, config):
        """对连接池进行初始化
        """
        pool_config = {}
        for section in config.prefix("dbpool_"):
            pool_name = section.split("_", 1)[1]
            pool_config[pool_name] = config[section]

            # 对字符串的配置项做一点微小的工作
            poolclass = pool_config[pool_name]["poolclass"]
            if isinstance(poolclass, str):
                if poolclass.lower() == "queuepool":
                    poolclass = QueuePool
                elif poolclass.lower() == "nullpool":
                    poolclass = NullPool
                else:
                    module_name, class_name = poolclass.rsplit(".", 1)
                    poolclass = getattr(__import__(module_name), class_name)
                pool_config[pool_name]["poolclass"] = poolclass

            for item_name in ("pool_size", "pool_recycle", "pool_timeout"):
                item_value = pool_config[pool_name][item_name]
                pool_config[pool_name][item_name] = int(item_value)
        return pool_config

    def _create_engine(self, config, section, all_pool_config):
        config_items = config[section]
        pool_policy = config[section]["pool_policy"]
        kws = {}
        if pool_policy is not None:
            pool_config = all_pool_config[pool_policy]
            if pool_config["poolclass"] != NullPool:
                kws = pool_config
        return create_engine(
            config_items["connect_url"],
            encoding=config_items["encoding"],
            connect_args=ujson.loads(config_items["connect_args"]),
            ** kws
        )

    def engine(self, engine_name):
        """根据引擎名字来获取引擎
        """
        if self._status != EngineManager.STATUS_OK:
            raise InvalidStatusException(u"Engine尚未初始化")
        return self._engines[engine_name]

    def dispose_all(self):
        """统一对引擎进行销毁
        """
        if self._status != EngineManager.STATUS_OK:
            return
        for _, engine_container in self._engines.items():
            engine_container.engine.dispose()
        self._status = EngineManager.STATUS_DISPOSED

_engine_manager = EngineManager()


class EngineContainer(object):

    """存储engine对象，并封装/代理一些与engine对象有关的操作
    """

    def __init__(self, engine):
        self.engine = engine
        self.sessionmaker = sessionmaker(bind=engine)
        self._base_model_class = None
        self._base_model_class_sem = threading.Lock()

    def session(self):
        return self.sessionmaker()

    @property
    def base_model_class(self):
        if self._base_model_class is not None:
            return self._base_model_class
        # 并发环境下，防止重复初始化
        with self._base_model_class_sem:
            # 虽然已经获得了锁，但是先人已经初始化了，立即返回
            if self._base_model_class is not None:
                return self._base_model_class
            self._base_model_class = automap_base()
            self._base_model_class.prepare(self.engine, reflect=True)
            return self._base_model_class


class OrmQueryPlugin(object):

    name = "orm_query"

    @staticmethod
    def config_validator(config):
        """配置验证器
        """
        global _engine_manager
        _engine_manager.validate_config(config)

    def __init__(self):
        self._engines = {}

    def sys_prepare(self, config):
        global _engine_manager
        _engine_manager.init_all(config)

    def execute(self, context, *exec_list):
        # 按顺序执行exec_list
        return [exec_(context) for exec_ in exec_list]

    def sys_cleanup(self, config):
        global _engine_manager
        _engine_manager.dispose_all()

_SELECT_STATEMENT_REGEX = re.compile("^select ", re.IGNORECASE)


class Query(object):

    """描述ORM查询信息以及执行查询操作
    """

    @args2fields()
    def __init__(self, engine_name, variable_name,
                 query_items, query=None, order_by=None, group_by=None,
                 params=None, row_handler=None, result_wrapper=None):
        """
        :param engine_name   使用的引擎名称
        :param variable_name 查询的结果将以此为变量名写入context
        :param query_items   查询项目，可以指定class类型和字符串类型
                             如果是字符串类型，那么会启用auto map进行处理
        :param query         接受回调函数或者是SQL以及字符串描述的查询条件
        :param params        如果基于文本或者SQL查询，那么可以通过此字段来传递参数
        :param row_handler   行处理器，针对每一行做格式转换操作
        :param result_wrapper 用于对查询结果进行包装，比如将查询结果包装成table对象
        """
        if isinstance(order_by, str):
            self._order_by = text(order_by)

    def __call__(self, context):
        global _engine_manager
        engine = _engine_manager.engine(self._engine_name)

        # 处理查询项
        query_items = []
        if isinstance(self._query_items, types.StringTypes):
            query_items.append(self.automap(engine, self._query_items))
        elif isinstance(self._query_items, collections.Iterable):
            for query_item in self._query_items:
                if isinstance(query_item, types.StringTypes):
                    query_item = self.automap(engine, query_item)
                query_items.append(query_item)
        else:
            query_items.append(self._query_items)

        # 解析params中的context变量
        if self._params:
            self._params = {
                key: parse_context_var(context, self._params[key])
                for key in self._params}

        session = engine.session()
        try:
            query = None
            if isinstance(self._query, types.StringTypes):
                if _SELECT_STATEMENT_REGEX.search(self._query):
                    if self._params is None:
                        self._params = {}
                    query = session.query(*query_items).from_statement(
                        text(self._query)).params(**self._params)
                else:
                    query = self._build_query(engine, session, query_items)
            elif isinstance(self._query, types.FunctionType):
                result = self._query(session, context, *query_items)
                # 可以返回查询对象，也可以返回查询结果
                if isinstance(result, ORMQuery):
                    query = result
                else:
                    if self._row_handler:
                        result = tuple(self._row_handler(row)
                                       for row in result)
                    if self._result_wrapper is not None:
                        result = self._result_wrapper(result)
                    context[self._variable_name] = result
                    return result
            else:
                query = self._build_query(engine, session, query_items)

            result = self._build_result(query)
            context[self._variable_name] = result
            return result
        finally:
            session.close()

    def automap(self, engine, query_item_str):
        base = engine.base_model_class
        if "." in query_item_str:
            table_name, field_name = query_item_str.split(".", 1)
            clazz = getattr(base.classes, table_name, None)
            return getattr(clazz, field_name, None)
        else:
            return getattr(base.classes, query_item_str, None)

    def _build_query(self, engine, session, query_items):
        query = session.query(*query_items)
        if self._query is not None:
            if isinstance(self._query, types.StringTypes):
                query = query.filter(text(self._query))
            else:
                query = query.filter(self._query)
        if self._order_by is not None:
            query = query.order_by(self._order_by)
        if self._group_by is not None:
            if isinstance(self._group_by, types.StringTypes):
                self._group_by = self.automap(engine, self._group_by)
            query = query.group_by(self._group_by)
        if self._params is not None:
            query = query.params(**self._params)
        return query

    def _build_result(self, query):
        if self._row_handler:
            result = tuple(self._row_handler(row) for row in query)
        else:
            result = query.all()
        if self._result_wrapper is not None:
            return self._result_wrapper(result)
        return result


class SQL(object):

    """使用该对象可以直接描述非ORM的SQL查询
    """

    @args2fields()
    def __init__(self, engine_name, variable_name, sql, params=None,
                 row_handler=None, result_wrapper=None):
        if params is None:
            self._params = {}

    def __call__(self, context):
        global _engine_manager

        engine_container = _engine_manager.engine(self._engine_name)
        session = engine_container.session()
        if isinstance(self._sql, types.StringTypes) and \
                _SELECT_STATEMENT_REGEX.search(self._sql):
            return self._execute_select_statement(session, context)
        else:
            # 非查询语句
            try:
                if isinstance(self._sql, types.StringTypes):
                    session.execute(self._sql, self._params)
                # 批量执行
                elif isinstance(self._sql, SequenceCollectionType):
                    if isinstance(self._params, SequenceCollectionType):
                        for idx, sql in enumerate(self._sql):
                            session.execute(sql, self._params[idx])
                    else:
                        for sql in self._sql:
                            session.execute(sql)
                session.commit()
            finally:
                session.close()

    def _execute_select_statement(self, session, context):
        try:
            result_proxy = session.execute(self._sql, self._params)
            result = None
            if self._row_handler is not None:
                result = tuple(self._row_handler(row) for row in result_proxy)
            else:
                result = tuple(tuple(row) for row in result_proxy)
            if self._result_wrapper is not None:
                result = self._result_wrapper(result)
            context[self._variable_name] = result
            result_proxy.close()
            return result
        finally:
            session.close()


class KeyExtractWrapper(object):

    """本Handler可以将一行中的某个字段转变为Key，并将结果包装为一个字典结构
       例如以第一列为Key：((1, 2, 3, 4, 5), ) => {1: (1, 2, 3, 4, 5)}
       以某个属性为Key: (User(id=0, name="SamChi", age=31),) => {"SamChi": User(...)}
    """

    def __init__(self, key_index):
        """
        :param key_index 作为索引，可以是数字，也可以是字符串属性名
                         具体由行类型来决定
        """
        self._key_index = key_index

    def __call__(self, query_result):
        if not query_result:
            return {}
        row_0 = query_result[0]
        if isinstance(row_0,
                      (types.ListType, types.TupleType, types.DictType)):
            return {row[self._key_index]: row for row in query_result}
        return {getattr(row, self._key_index): row for row in query_result}



# coding: utf-8

"""内置模板集中处
"""

TEMPLATE_TEMPLATE = """
# coding: utf-8

\"\"\"生成模板的模板
\"\"\"

import argparse
from girlfriend.util import file_template, script

parser = argparse.ArgumentParser(description=__doc__.decode("utf-8"))
parser.add_argument("-n", dest="name",
                    action="store", help=u"模板名称")


CONTENT = \"\"\"
# coding: utf-8

\\"\\"\\"
Docs goes here
\\"\\"\\"

import argparse
from girlfriend.util.file_template import Dir, File
from girlfriend.util.script import (
    show_traceback_and_exit,
    show_msg_and_exit
)

parser = argparse.ArgumentParser(description=__doc__.decode("utf-8"))
parser.add_argument("--n", dest="name",
                    action="store", help=u"模板名称")


def gen(path, name):
    pass
\"\"\"


def gen(name, path):
    if not name:
        script.show_msg_and_exit(u"必须使用-n参数来指定模板文件名")
    if not name.endswith(".py"):
        name += ".py"
    file_template.File(name, content=CONTENT).makeme(path)
"""



# coding: utf-8

"""针对config对象的测试用例
"""

import fixtures
from girlfriend.testing import GirlFriendTestCase
from girlfriend.util.config import Config


class ConfigTestCase(GirlFriendTestCase):

    def test_getsetattr(self):
        cfg = Config()
        cfg["a"] = 1
        self.assertEquals(cfg.a, 1)
        cfg.a = 2
        self.assertEquals(cfg.a, 2)
        cfg.b = Config()
        cfg.b.a = 3
        self.assertEquals(cfg.b.a, 3)


class ConfigLoadFixture(fixtures.Fixture):

    """加载默认配置文件 $HOME/.gf/gf.config
    """

    def setUp(self):
        self.config = Config.load_by_name("default")



# coding: utf-8

"""workflow.protocol模块的测试用例
"""

from girlfriend.testing import GirlFriendTestCase
from girlfriend.workflow.protocol import AbstractListener


class AbstractListenerTestCase(GirlFriendTestCase):

    """AbstractListener的测试用例
    """

    def test_wrap_functions(self):

        listener = AbstractListener.wrap_function((
            "on_start", lambda ctx: "on_start",
            "unit_start", lambda ctx: "on_unit_start",
            "unit_finish", lambda ctx: "on_unit_finish",
            "on_finish", lambda ctx: "on_finish",
        ))

        self.assertEquals(listener.on_start(None), "on_start")
        self.assertEquals(listener.on_finish(None), "on_finish")
        self.assertEquals(listener.on_unit_start(None), "on_unit_start")
        self.assertEquals(listener.on_unit_finish(None), "on_unit_finish")
